"""
Embeddings module
"""

import json
import os
import tempfile

from ..ann import ANNFactory
from ..archive import ArchiveFactory
from ..cloud import CloudFactory
from ..database import DatabaseFactory
from ..graph import GraphFactory
from ..scoring import ScoringFactory
from ..vectors import VectorsFactory

from .index import Action, Configuration, Functions, Indexes, IndexIds, Reducer, Stream, Transform
from .search import Explain, Ids, Query, Search, Terms


# pylint: disable=C0302,R0904
class Embeddings:
    """
    Embeddings databases are the engine that delivers semantic search. Data is transformed into embeddings vectors where similar concepts
    will produce similar vectors. Indexes both large and small are built with these vectors. The indexes are used to find results
    that have the same meaning, not necessarily the same keywords.
    """

    # pylint: disable = W0231
    def __init__(self, config=None, models=None, **kwargs):
        """
        Creates a new embeddings index. Embeddings indexes are thread-safe for read operations but writes must be synchronized.

        Args:
            config: embeddings configuration
            models: models cache, used for model sharing between embeddings
            kwargs: additional configuration as keyword args
        """

        # Index configuration
        self.config = None

        # Dimensionality reduction - word vectors only
        self.reducer = None

        # Dense vector model - transforms data into similarity vectors
        self.model = None

        # Approximate nearest neighbor index
        self.ann = None

        # Index ids when content is disabled
        self.ids = None

        # Document database
        self.database = None

        # Resolvable functions
        self.functions = None

        # Graph network
        self.graph = None

        # Sparse vectors
        self.scoring = None

        # Query model
        self.query = None

        # Index archive
        self.archive = None

        # Subindexes for this embeddings instance
        self.indexes = None

        # Models cache
        self.models = models

        # Merge configuration into single dictionary
        config = {**config, **kwargs} if config and kwargs else kwargs if kwargs else config

        # Set initial configuration
        self.configure(config)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def score(self, documents):
        """
        Builds a term weighting scoring index. Only used by word vectors models.

        Args:
            documents: iterable of (id, data, tags), (id, data) or data
        """

        # Build scoring index for word vectors term weighting
        if self.isweighted():
            self.scoring.index(Stream(self)(documents))

    def index(self, documents, reindex=False, checkpoint=None):
        """
        Builds an embeddings index. This method overwrites an existing index.

        Args:
            documents: iterable of (id, data, tags), (id, data) or data
            reindex: if this is a reindex operation in which case database creation is skipped, defaults to False
            checkpoint: optional checkpoint directory, enables indexing restart
        """

        # Initialize index
        self.initindex(reindex)

        # Create transform and stream
        transform = Transform(self, Action.REINDEX if reindex else Action.INDEX, checkpoint)
        stream = Stream(self, Action.REINDEX if reindex else Action.INDEX)

        with tempfile.NamedTemporaryFile(mode="wb", suffix=".npy") as buffer:
            # Load documents into database and transform to vectors
            ids, dimensions, embeddings = transform(stream(documents), buffer)
            if embeddings is not None:
                # Build LSA model (if enabled). Remove principal components from embeddings.
                if self.config.get("pca"):
                    self.reducer = Reducer(embeddings, self.config["pca"])
                    self.reducer(embeddings)

                # Save index dimensions
                self.config["dimensions"] = dimensions

                # Create approximate nearest neighbor index
                self.ann = self.createann()

                # Add embeddings to the index
                self.ann.index(embeddings)

            # Save indexids-ids mapping for indexes with no database, except when this is a reindex
            if ids and not reindex and not self.database:
                self.ids = self.createids(ids)

        # Index scoring, if necessary
        # This must occur before graph index in order to be available to the graph
        if self.issparse():
            self.scoring.index()

        # Index subindexes, if necessary
        if self.indexes:
            self.indexes.index()

        # Index graph, if necessary
        if self.graph:
            self.graph.index(Search(self, indexonly=True), Ids(self), self.batchsimilarity)

    def upsert(self, documents, checkpoint=None):
        """
        Runs an embeddings upsert operation. If the index exists, new data is
        appended to the index, existing data is updated. If the index doesn't exist,
        this method runs a standard index operation.

        Args:
            documents: iterable of (id, data, tags), (id, data) or data
            checkpoint: optional checkpoint directory, enables indexing restart
        """

        # Run standard insert if index doesn't exist or it has no records
        if not self.count():
            self.index(documents, checkpoint=checkpoint)
            return

        # Create transform and stream
        transform = Transform(self, Action.UPSERT, checkpoint=checkpoint)
        stream = Stream(self, Action.UPSERT)

        with tempfile.NamedTemporaryFile(mode="wb", suffix=".npy") as buffer:
            # Load documents into database and transform to vectors
            ids, _, embeddings = transform(stream(documents), buffer)
            if embeddings is not None:
                # Remove principal components from embeddings, if necessary
                if self.reducer:
                    self.reducer(embeddings)

                # Append embeddings to the index
                self.ann.append(embeddings)

            # Save indexids-ids mapping for indexes with no database
            if ids and not self.database:
                self.ids = self.createids(self.ids + ids)

        # Scoring upsert, if necessary
        # This must occur before graph upsert in order to be available to the graph
        if self.issparse():
            self.scoring.upsert()

        # Subindexes upsert, if necessary
        if self.indexes:
            self.indexes.upsert()

        # Graph upsert, if necessary
        if self.graph:
            self.graph.upsert(Search(self, indexonly=True), Ids(self), self.batchsimilarity)

    def delete(self, ids):
        """
        Deletes from an embeddings index. Returns list of ids deleted.

        Args:
            ids: list of ids to delete

        Returns:
            list of ids deleted
        """

        # List of internal indices for each candidate id to delete
        indices = []

        # List of deleted ids
        deletes = []

        if self.database:
            # Retrieve indexid-id mappings from database
            ids = self.database.ids(ids)

            # Parse out indices and ids to delete
            indices = [i for i, _ in ids]
            deletes = sorted(set(uid for _, uid in ids))

            # Delete ids from database
            self.database.delete(deletes)
        elif self.ann or self.scoring:
            # Find existing ids
            for uid in ids:
                indices.extend([index for index, value in enumerate(self.ids) if uid == value])

            # Clear embeddings ids
            for index in indices:
                deletes.append(self.ids[index])
                self.ids[index] = None

        # Delete indices for all indexes and data stores
        if indices:
            # Delete ids from ann
            if self.isdense():
                self.ann.delete(indices)

            # Delete ids from scoring
            if self.issparse():
                self.scoring.delete(indices)

            # Delete ids from subindexes
            if self.indexes:
                self.indexes.delete(indices)

            # Delete ids from graph
            if self.graph:
                self.graph.delete(indices)

        return deletes

    def reindex(self, config=None, function=None, **kwargs):
        """
        Recreates embeddings index using config. This method only works if document content storage is enabled.

        Args:
            config: new config
            function: optional function to prepare content for indexing
            kwargs: additional configuration as keyword args
        """

        if self.database:
            # Merge configuration into single dictionary
            config = {**config, **kwargs} if config and kwargs else config if config else kwargs

            # Keep content and objects parameters to ensure database is preserved
            config["content"] = self.config["content"]
            if "objects" in self.config:
                config["objects"] = self.config["objects"]

            # Reset configuration
            self.configure(config)

            # Reset function references
            if self.functions:
                self.functions.reset()

            # Reindex
            if function:
                self.index(function(self.database.reindex(self.config)), True)
            else:
                self.index(self.database.reindex(self.config), True)

    def transform(self, document, category=None, index=None):
        """
        Transforms document into an embeddings vector.

        Args:
            documents: iterable of (id, data, tags), (id, data) or data
            category: category for instruction-based embeddings
            index: index name, if applicable

        Returns:
            embeddings vector
        """

        return self.batchtransform([document], category, index)[0]

    def batchtransform(self, documents, category=None, index=None):
        """
        Transforms documents into embeddings vectors.

        Args:
            documents: iterable of (id, data, tags), (id, data) or data
            category: category for instruction-based embeddings
            index: index name, if applicable

        Returns:
            embeddings vectors
        """

        # Initialize default parameters, if necessary
        self.defaults()

        # Get vector model
        model = self.findmodel(index)

        # Convert documents into embeddings
        embeddings = model.batchtransform(Stream(self)(documents), category)

        # Reduce the dimensionality of the embeddings. Scale the embeddings using this
        # model to reduce the noise of common but less relevant terms.
        if self.reducer:
            self.reducer(embeddings)

        return embeddings

    def count(self):
        """
        Total number of elements in this embeddings index.

        Returns:
            number of elements in this embeddings index
        """

        if self.ann:
            return self.ann.count()
        if self.scoring:
            return self.scoring.count()
        if self.database:
            return self.database.count()
        if self.ids:
            return len([uid for uid in self.ids if uid is not None])

        # Default to 0 when no suitable method found
        return 0

    def search(self, query, limit=None, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input query. This method runs an index search, index + database search
        or a graph search, depending on the embeddings configuration and query.

        Args:
            query: input query
            limit: maximum results
            weights: hybrid score weights, if applicable
            index: index name, if applicable
            parameters: dict of named parameters to bind to placeholders
            graph: return graph results if True

        Returns:
            list of (id, score) for index search
            list of dict for an index + database search
            graph when graph is set to True
        """

        results = self.batchsearch([query], limit, weights, index, [parameters], graph)
        return results[0] if results else results

    def batchsearch(self, queries, limit=None, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input query. This method runs an index search, index + database search
        or a graph search, depending on the embeddings configuration and query.

        Args:
            queries: input queries
            limit: maximum results
            weights: hybrid score weights, if applicable
            index: index name, if applicable
            parameters: list of dicts of named parameters to bind to placeholders
            graph: return graph results if True

        Returns:
            list of (id, score) per query for index search
            list of dict per query for an index + database search
            list of graph per query when graph is set to True
        """

        # Determine if graphs should be returned
        graph = graph if self.graph else False

        # Execute search
        results = Search(self, indexids=graph)(queries, limit, weights, index, parameters)

        # Create subgraphs using results, if necessary
        return [self.graph.filter(x) if isinstance(x, list) else x for x in results] if graph else results

    def similarity(self, query, data):
        """
        Computes the similarity between query and list of data. Returns a list of
        (id, score) sorted by highest score, where id is the index in data.

        Args:
            query: input query
            data: list of data

        Returns:
            list of (id, score)
        """

        return self.batchsimilarity([query], data)[0]

    def batchsimilarity(self, queries, data):
        """
        Computes the similarity between list of queries and list of data. Returns a list
        of (id, score) sorted by highest score per query, where id is the index in data.

        Args:
            queries: input queries
            data: list of data

        Returns:
            list of (id, score) per query
        """

        # Convert queries to embedding vectors
        queries = self.batchtransform(((None, query, None) for query in queries), "query")
        data = self.batchtransform(((None, row, None) for row in data), "data")

        # Get vector model
        model = self.findmodel()

        # Dot product on normalized vectors is equal to cosine similarity
        scores = model.dot(queries, data)

        # Add index and sort desc based on score
        return [sorted(enumerate(score), key=lambda x: x[1], reverse=True) for score in scores]

    def explain(self, query, texts=None, limit=None):
        """
        Explains the importance of each input token in text for a query. This method requires either content to be enabled
        or texts to be provided.

        Args:
            query: input query
            texts: optional list of (text|list of tokens), otherwise runs search query
            limit: optional limit if texts is None

        Returns:
            list of dict per input text where a higher token scores represents higher importance relative to the query
        """

        results = self.batchexplain([query], texts, limit)
        return results[0] if results else results

    def batchexplain(self, queries, texts=None, limit=None):
        """
        Explains the importance of each input token in text for a list of queries. This method requires either content to be enabled
        or texts to be provided.

        Args:
            queries: input queries
            texts: optional list of (text|list of tokens), otherwise runs search queries
            limit: optional limit if texts is None

        Returns:
            list of dict per input text per query where a higher token scores represents higher importance relative to the query
        """

        return Explain(self)(queries, texts, limit)

    def terms(self, query):
        """
        Extracts keyword terms from a query.

        Args:
            query: input query

        Returns:
            query reduced down to keyword terms
        """

        return self.batchterms([query])[0]

    def batchterms(self, queries):
        """
        Extracts keyword terms from a list of queries.

        Args:
            queries: list of queries

        Returns:
            list of queries reduced down to keyword term strings
        """

        return Terms(self)(queries)

    def exists(self, path=None, cloud=None, **kwargs):
        """
        Checks if an index exists at path.

        Args:
            path: input path
            cloud: cloud storage configuration
            kwargs: additional configuration as keyword args

        Returns:
            True if index exists, False otherwise
        """

        # Check if this exists in a cloud instance
        cloud = self.createcloud(cloud=cloud, **kwargs)
        if cloud:
            return cloud.exists(path)

        # Check if this is an archive file and exists
        path, apath = self.checkarchive(path)
        if apath:
            return os.path.exists(apath)

        # Return true if path has a config.json or config file with an offset set
        return path and (os.path.exists(f"{path}/config.json") or os.path.exists(f"{path}/config")) and "offset" in Configuration().load(path)

    def load(self, path=None, cloud=None, config=None, **kwargs):
        """
        Loads an existing index from path.

        Args:
            path: input path
            cloud: cloud storage configuration
            config: configuration overrides
            kwargs: additional configuration as keyword args

        Returns:
            Embeddings
        """

        # Load from cloud, if configured
        cloud = self.createcloud(cloud=cloud, **kwargs)
        if cloud:
            path = cloud.load(path)

        # Check if this is an archive file and extract
        path, apath = self.checkarchive(path)
        if apath:
            self.archive.load(apath)

        # Load index configuration
        self.config = Configuration().load(path)

        # Apply config overrides
        self.config = {**self.config, **config} if config else self.config

        # Approximate nearest neighbor index - stores dense vectors
        self.ann = self.createann()
        if self.ann:
            self.ann.load(f"{path}/embeddings")

        # Dimensionality reduction model - word vectors only
        if self.config.get("pca"):
            self.reducer = Reducer()
            self.reducer.load(f"{path}/lsa")

        # Index ids when content is disabled
        self.ids = self.createids()
        if self.ids:
            self.ids.load(f"{path}/ids")

        # Document database - stores document content
        self.database = self.createdatabase()
        if self.database:
            self.database.load(f"{path}/documents")

        # Sparse vectors - stores term sparse arrays
        self.scoring = self.createscoring()
        if self.scoring:
            self.scoring.load(f"{path}/scoring")

        # Subindexes
        self.indexes = self.createindexes()
        if self.indexes:
            self.indexes.load(f"{path}/indexes")

        # Graph network - stores relationships
        self.graph = self.creategraph()
        if self.graph:
            self.graph.load(f"{path}/graph")

        # Dense vectors - transforms data to embeddings vectors
        self.model = self.loadvectors()

        # Query model
        self.query = self.loadquery()

        return self

    def save(self, path, cloud=None, **kwargs):
        """
        Saves an index in a directory at path unless path ends with tar.gz, tar.bz2, tar.xz or zip.
        In those cases, the index is stored as a compressed file.

        Args:
            path: output path
            cloud: cloud storage configuration
            kwargs: additional configuration as keyword args
        """

        if self.config:
            # Check if this is an archive file
            path, apath = self.checkarchive(path)

            # Create output directory, if necessary
            os.makedirs(path, exist_ok=True)

            # Save index configuration
            Configuration().save(self.config, path)

            # Save approximate nearest neighbor index
            if self.ann:
                self.ann.save(f"{path}/embeddings")

            # Save dimensionality reduction model (word vectors only)
            if self.reducer:
                self.reducer.save(f"{path}/lsa")

            # Save index ids
            if self.ids:
                self.ids.save(f"{path}/ids")

            # Save document database
            if self.database:
                self.database.save(f"{path}/documents")

            # Save scoring index
            if self.scoring:
                self.scoring.save(f"{path}/scoring")

            # Save subindexes
            if self.indexes:
                self.indexes.save(f"{path}/indexes")

            # Save graph
            if self.graph:
                self.graph.save(f"{path}/graph")

            # If this is an archive, save it
            if apath:
                self.archive.save(apath)

            # Save to cloud, if configured
            cloud = self.createcloud(cloud=cloud, **kwargs)
            if cloud:
                cloud.save(apath if apath else path)

    def close(self):
        """
        Closes this embeddings index and frees all resources.
        """

        self.config, self.archive = None, None
        self.reducer, self.query = None, None
        self.ids = None

        # Close ANN
        if self.ann:
            self.ann.close()
            self.ann = None

        # Close database
        if self.database:
            self.database.close()
            self.database, self.functions = None, None

        # Close scoring
        if self.scoring:
            self.scoring.close()
            self.scoring = None

        # Close graph
        if self.graph:
            self.graph.close()
            self.graph = None

        # Close indexes
        if self.indexes:
            self.indexes.close()
            self.indexes = None

        # Close vectors model
        if self.model:
            self.model.close()
            self.model = None

        self.models = None

    def info(self):
        """
        Prints the current embeddings index configuration.
        """

        if self.config:
            # Print configuration
            print(json.dumps(self.config, sort_keys=True, default=str, indent=2))

    def issparse(self):
        """
        Checks if this instance has an associated sparse keyword or sparse vectors scoring index.

        Returns:
            True if scoring has an associated sparse keyword/vector index, False otherwise
        """

        return self.scoring and self.scoring.issparse()

    def isdense(self):
        """
        Checks if this instance has an associated ANN instance.

        Returns:
            True if this instance has an associated ANN, False otherwise
        """

        return self.ann is not None

    def isweighted(self):
        """
        Checks if this instance has an associated scoring instance with term weighting enabled.

        Returns:
            True if term weighting is enabled, False otherwise
        """

        return self.scoring and self.scoring.isweighted()

    def findmodel(self, index=None):
        """
        Finds the primary vector model used by this instance.

        Returns:
            Vectors
        """

        return (
            self.indexes.findmodel(index)
            if index and self.indexes
            else (
                self.model
                if self.model
                else self.scoring.findmodel() if self.scoring and self.scoring.findmodel() else self.indexes.findmodel() if self.indexes else None
            )
        )

    def configure(self, config):
        """
        Sets the configuration for this embeddings index and loads config-driven models.

        Args:
            config: embeddings configuration
        """

        # Configuration
        self.config = config

        # Dimensionality reduction model
        self.reducer = None

        # Create scoring instance for word vectors term weighting
        scoring = self.config.get("scoring") if self.config else None
        self.scoring = self.createscoring() if scoring and not self.hassparse() else None

        # Dense vectors - transforms data to embeddings vectors
        self.model = self.loadvectors() if self.config else None

        # Query model
        self.query = self.loadquery() if self.config else None

    def initindex(self, reindex):
        """
        Initialize new index.

        Args:
            reindex: if this is a reindex operation in which case database creation is skipped, defaults to False
        """

        # Initialize default parameters, if necessary
        self.defaults()

        # Initialize index ids, only created when content is disabled
        self.ids = None

        # Create document database, if necessary
        if not reindex:
            self.database = self.createdatabase()

            # Reset archive since this is a new index
            self.archive = None

        # Close existing ANN, if necessary
        if self.ann:
            self.ann.close()

        # Initialize ANN, will be created after index transformations complete
        self.ann = None

        # Create scoring only if the scoring config is for a sparse index
        if self.hassparse():
            self.scoring = self.createscoring()

        # Create subindexes, if necessary
        self.indexes = self.createindexes()

        # Create graph, if necessary
        self.graph = self.creategraph()

    def defaults(self):
        """
        Apply default parameters to current configuration.

        Returns:
            configuration with default parameters set
        """

        self.config = self.config if self.config else {}

        # Expand sparse index shortcuts
        if not self.config.get("scoring") and any(self.config.get(key) for key in ["keyword", "sparse", "hybrid"]):
            self.defaultsparse()

        # Expand graph shortcuts
        if self.config.get("graph") is True:
            self.config["graph"] = {}

        # Check if default model should be loaded
        if not self.model and (self.defaultallowed() or self.config.get("dense")):
            self.config["path"] = "sentence-transformers/all-MiniLM-L6-v2"

            # Load dense vectors model
            self.model = self.loadvectors()

    def defaultsparse(self):
        """
        Logic to derive default sparse index configuration.
        """

        # Check for keyword and hybrid parameters
        method = None
        for x in ["keyword", "hybrid"]:
            value = self.config.get(x)
            if value:
                method = value if isinstance(value, str) else "bm25"

                # Enable dense index when hybrid enabled
                if x == "hybrid":
                    self.config["dense"] = True

        sparse = self.config.get("sparse", {})
        if sparse or method == "sparse":
            # Sparse vector configuration
            sparse = {"path": self.config.get("sparse")} if isinstance(sparse, str) else {} if isinstance(sparse, bool) else sparse
            sparse["path"] = sparse.get("path", "opensearch-project/opensearch-neural-sparse-encoding-doc-v2-mini")

            # Merge in sparse parameters
            self.config["scoring"] = {**{"method": "sparse"}, **sparse}

        elif method:
            # Sparse keyword configuration
            self.config["scoring"] = {"method": method, "terms": True, "normalize": True}

    def defaultallowed(self):
        """
        Tests if this embeddings instance can use a default model if not otherwise provided.

        Returns:
            True if a default model is allowed, False otherwise
        """

        params = [("keyword", False), ("sparse", False), ("defaults", True)]
        return all(self.config.get(key, default) == default for key, default in params)

    def loadvectors(self):
        """
        Loads a vector model set in config.

        Returns:
            vector model
        """

        # Create model cache if subindexes are enabled
        if "indexes" in self.config and self.models is None:
            self.models = {}

        # Support path via dense parameter
        dense = self.config.get("dense")
        if not self.config.get("path") and dense and isinstance(dense, str):
            self.config["path"] = dense

        # Load vector model
        return VectorsFactory.create(self.config, self.scoring, self.models)

    def loadquery(self):
        """
        Loads a query model set in config.

        Returns:
            query model
        """

        if "query" in self.config:
            return Query(**self.config["query"])

        return None

    def checkarchive(self, path):
        """
        Checks if path is an archive file.

        Args:
            path: path to check

        Returns:
            (working directory, current path) if this is an archive, original path otherwise
        """

        # Create archive instance, if necessary
        self.archive = ArchiveFactory.create()

        # Check if path is an archive file
        if self.archive.isarchive(path):
            # Return temporary archive working directory and original path
            return self.archive.path(), path

        return path, None

    def createcloud(self, **cloud):
        """
        Creates a cloud instance from config.

        Args:
            cloud: cloud configuration
        """

        # Merge keyword args and keys under the cloud parameter
        config = cloud
        if "cloud" in config and config["cloud"]:
            config.update(config.pop("cloud"))

        # Create cloud instance from config and return
        return CloudFactory.create(config) if config else None

    def createann(self):
        """
        Creates an ANN from config.

        Returns:
            new ANN, if enabled in config
        """

        # Free existing resources
        if self.ann:
            self.ann.close()

        return ANNFactory.create(self.config) if self.config.get("path") or self.defaultallowed() else None

    def createdatabase(self):
        """
        Creates a database from config. This method will also close any existing database connection.

        Returns:
            new database, if enabled in config
        """

        # Free existing resources
        if self.database:
            self.database.close()

        config = self.config.copy()

        # Create references to callable functions
        self.functions = Functions(self) if "functions" in config else None
        if self.functions:
            config["functions"] = self.functions(config)

        # Create database from config and return
        return DatabaseFactory.create(config)

    def creategraph(self):
        """
        Creates a graph from config.

        Returns:
            new graph, if enabled in config
        """

        # Free existing resources
        if self.graph:
            self.graph.close()

        if "graph" in self.config:
            # Get or create graph configuration
            config = self.config["graph"] if "graph" in self.config else {}

            # Create configuration with custom columns, if necessary
            config = self.columns(config)
            return GraphFactory.create(config)

        return None

    def createids(self, ids=None):
        """
        Creates indexids when content is disabled.

        Args:
            ids: optional ids to add

        Returns:
            new indexids, if content disabled
        """

        # Load index ids when content is disabled
        return IndexIds(self, ids) if not self.config.get("content") else None

    def createindexes(self):
        """
        Creates subindexes from config.

        Returns:
            list of subindexes
        """

        # Free existing resources
        if self.indexes:
            self.indexes.close()

        # Load subindexes
        if "indexes" in self.config:
            indexes = {}
            for index, config in self.config["indexes"].items():
                # Create index with shared model cache
                indexes[index] = Embeddings(config, models=self.models)

            # Wrap as Indexes object
            return Indexes(self, indexes)

        return None

    def createscoring(self):
        """
        Creates a scoring from config.

        Returns:
            new scoring, if enabled in config
        """

        # Free existing resources
        if self.scoring:
            self.scoring.close()

        if "scoring" in self.config:
            # Expand scoring to a dictionary, if necessary
            config = self.config["scoring"]
            config = config if isinstance(config, dict) else {"method": config}

            # Create configuration with custom columns, if necessary
            config = self.columns(config)
            return ScoringFactory.create(config, self.models)

        return None

    def hassparse(self):
        """
        Checks is this embeddings database has an associated sparse index.

        Returns:
            True if this embeddings has an associated scoring index
        """

        # Create scoring only if scoring is a sparse keyword/vector index
        return ScoringFactory.issparse(self.config.get("scoring"))

    def columns(self, config):
        """
        Adds custom text/object column information if it's provided.

        Args:
            config: input configuration

        Returns:
            config with column information added
        """

        # Add text/object columns if custom
        if "columns" in self.config:
            # Work on copy of configuration
            config = config.copy()

            # Copy columns to config
            config["columns"] = self.config["columns"]

        return config



"""
Factory module
"""

from ...util import Resolver

from .sbert import SparseSTVectors


class SparseVectorsFactory:
    """
    Methods to create sparse vector models.
    """

    @staticmethod
    def create(config, models=None):
        """
        Create a Vectors model instance.

        Args:
            config: vector configuration
            models: models cache

        Returns:
            Vectors
        """

        # Get vector method
        method = config.get("method", "sentence-transformers")

        # Sentence Transformers vectors
        if method == "sentence-transformers":
            return SparseSTVectors(config, None, models) if config and config.get("path") else None

        # Resolve custom method
        return SparseVectorsFactory.resolve(method, config, models) if method else None

    @staticmethod
    def resolve(backend, config, models):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: vector configuration
            models: models cache

        Returns:
            Vectors
        """

        try:
            return Resolver()(backend)(config, None, models)
        except Exception as e:
            raise ImportError(f"Unable to resolve sparse vectors backend: '{backend}'") from e



"""
PGSparse module
"""

import os

import numpy as np

# Conditional import
try:
    from pgvector import SparseVector
    from pgvector.sqlalchemy import SPARSEVEC

    PGSPARSE = True
except ImportError:
    PGSPARSE = False

from ..dense import PGVector


class PGSparse(PGVector):
    """
    Builds a Sparse ANN index backed by a Postgres database.
    """

    def __init__(self, config):
        if not PGSPARSE:
            raise ImportError('PGSparse is not available - install "ann" extra to enable')

        super().__init__(config)

        # Quantization not supported
        self.qbits = None

    def defaulttable(self):
        return "svectors"

    def url(self):
        return self.setting("url", os.environ.get("SCORING_URL", os.environ.get("ANN_URL")))

    def column(self):
        return SPARSEVEC(self.config["dimensions"])

    def operation(self):
        return "sparsevec_ip_ops"

    def prepare(self, data):
        # pgvector only allows 1000 non-zero values for sparse vectors
        # Trim to top 1000 values, if necessary
        if data.count_nonzero() > 1000:
            value = -np.sort(-data[0, :].data)[1000]
            data.data = np.where(data.data > value, data.data, 0)
            data.eliminate_zeros()

        # Wrap as sparse vector
        return SparseVector(data)



"""
Index imports
"""

from .action import Action
from .autoid import AutoId
from .configuration import Configuration
from .documents import Documents
from .functions import Functions
from .indexes import Indexes
from .indexids import IndexIds
from .reducer import Reducer
from .stream import Stream
from .transform import Transform



"""
Objects module
"""

# Conditional import
try:
    from PIL import Image

    PIL = True
except ImportError:
    PIL = False

from ..hfpipeline import HFPipeline


class Objects(HFPipeline):
    """
    Applies object detection models to images. Supports both object detection models and image classification models.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, classification=False, threshold=0.9, **kwargs):
        if not PIL:
            raise ImportError('Objects pipeline is not available - install "pipeline" extra to enable')

        super().__init__("image-classification" if classification else "object-detection", path, quantize, gpu, model, **kwargs)

        self.classification = classification
        self.threshold = threshold

    def __call__(self, images, flatten=False, workers=0):
        """
        Applies object detection/image classification models to images. Returns a list of (label, score).

        This method supports a single image or a list of images. If the input is an image, the return
        type is a 1D list of (label, score). If text is a list, a 2D list of (label, score) is
        returned with a row per image.

        Args:
            images: image|list
            flatten: flatten output to a list of objects
            workers: number of concurrent workers to use for processing data, defaults to None

        Returns:
            list of (label, score)
        """

        # Convert single element to list
        values = [images] if not isinstance(images, list) else images

        # Open images if file strings
        values = [Image.open(image) if isinstance(image, str) else image for image in values]

        # Run pipeline
        results = (
            self.pipeline(values, num_workers=workers)
            if self.classification
            else self.pipeline(values, threshold=self.threshold, num_workers=workers)
        )

        # Build list of (id, score)
        outputs = []
        for result in results:
            # Convert to (label, score) tuples
            result = [(x["label"], x["score"]) for x in result if x["score"] > self.threshold]

            # Sort by score descending
            result = sorted(result, key=lambda x: x[1], reverse=True)

            # Deduplicate labels
            unique = set()
            elements = []
            for label, score in result:
                if label not in unique:
                    elements.append(label if flatten else (label, score))
                    unique.add(label)

            outputs.append(elements)

        # Return single element if single element passed in
        return outputs[0] if not isinstance(images, list) else outputs



"""
Defines API paths for workflow endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.post("/workflow")
def workflow(name: str = Body(...), elements: List = Body(...)):
    """
    Executes a named workflow using elements as input.

    Args:
        name: workflow name
        elements: list of elements to run through workflow

    Returns:
        list of processed elements
    """

    return application.get().workflow(name, elements)



"""
Token module
"""


class Token:
    """
    Methods to check for token type.
    """

    # Similar token replacement
    SIMILAR_TOKEN = "__SIMILAR__"

    # Default distinct token
    DISTINCT = ["distinct"]

    # Default alias token
    ALIAS = ["as"]

    # Default list of comparison operators
    OPERATORS = ["=", "!=", "<>", ">", ">=", "<", "<=", "+", "-", "*", "/", "%", "||", "not", "between", "like", "is", "null"]

    # Default list of logic separators
    LOGIC_SEPARATORS = ["and", "or"]

    # Default list of sort order operators
    SORT_ORDER = ["asc", "desc"]

    @staticmethod
    def get(tokens, x):
        """
        Gets token at position x. This method will validate position is valid within tokens.

        Args:
            tokens: input tokens
            x: position to retrieve

        Returns:
            tokens[x] if x is a valid position, None otherwise
        """

        if 0 <= x < len(tokens):
            return tokens[x]

        return None

    @staticmethod
    def isalias(tokens, x, alias):
        """
        Checks if tokens[x] is an alias keyword.

        Args:
            tokens: input tokens
            x: current position
            alias: if column alias processing is enabled

        Returns:
            True if tokens[x] is an alias token, False otherwise
        """

        prior = Token.get(tokens, x - 1)
        token = tokens[x]

        # True if prior token is not a separator, grouping token or distinct token and current token is either a column token or quoted token
        return (
            alias
            and x > 0
            and not Token.isseparator(prior)
            and not Token.isgroupstart(prior)
            and not Token.isdistinct(prior)
            and (Token.iscolumn(token) or Token.isquoted(token))
        )

    @staticmethod
    def isattribute(tokens, x):
        """
        Checks if tokens[x] is an attribute.

        Args:
            tokens: input tokens
            x: current position

        Returns:
            True if tokens[x] is an attribute, False otherwise
        """

        # True if token is a column and next token is not an operator
        return Token.iscolumn(tokens[x]) and not Token.isoperator(Token.get(tokens, x + 1))

    @staticmethod
    def isbracket(token):
        """
        Checks if token is an open bracket.

        Args:
            token: token to test

        Returns:
            True if token is an open bracket, False otherwise
        """

        # Token is a bracket
        return token == "["

    @staticmethod
    def iscolumn(token):
        """
        Checks if token is a column name.

        Args:
            token: token to test

        Returns:
            True if this token is a column name token, False otherwise
        """

        # Columns are not operators, logic separators, literals or sort order tokens
        return (
            token
            and not Token.isoperator(token)
            and not Token.islogicseparator(token)
            and not Token.isliteral(token)
            and not Token.issortorder(token)
        )

    @staticmethod
    def iscompound(tokens, x):
        """
        Checks if tokens[x] is a compound expression.

        Args:
            tokens: input tokens
            x: current position

        Returns:
            True if tokens[x] is a compound expression, False otherwise
        """

        # Compound expression is defined as: <column> <operator(s)> <column>
        return Token.isoperator(tokens[x]) and (Token.iscolumn(Token.get(tokens, x - 1)) or Token.iscolumn(Token.get(tokens, x + 1)))

    @staticmethod
    def isdistinct(token):
        """
        Checks if token is the distinct keyword.

        Args:
            token: token to test

        Returns:
            True if this token is a distinct keyword, False otherwise
        """

        # Token is the distinct keyword
        return token and token.lower() in Token.DISTINCT

    @staticmethod
    def isfunction(tokens, x):
        """
        Checks if tokens[x] is a function.

        Args:
            tokens: input tokens
            x: current position

        Returns:
            True if tokens[x] is a function, False otherwise
        """

        # True if a column token is followed by an open paren
        return Token.iscolumn(tokens[x]) and Token.get(tokens, x + 1) == "("

    @staticmethod
    def isgroupstart(token):
        """
        Checks if token is a group start token.

        Args:
            token: token to test

        Returns:
            True if token is a group start token, False otherwise
        """

        # Token is a paren
        return token == "("

    @staticmethod
    def isliteral(token):
        """
        Checks if token is a literal.

        Args:
            token: token to test

        Returns:
            True if this token is a literal, False otherwise
        """

        # Literals are wrapped in quotes, parens, wildcards or numeric.
        return token and (token.startswith(("'", '"', ",", "(", ")", "*")) or token.replace(".", "", 1).isdigit())

    @staticmethod
    def islogicseparator(token):
        """
        Checks if token is a logic separator token.

        Args:
            token: token to test

        Returns:
            True if this token is a logic separator, False otherwise
        """

        # Token is a logic separator
        return token and token.lower() in Token.LOGIC_SEPARATORS

    @staticmethod
    def isoperator(token):
        """
        Checks if token is an operator token.

        Args:
            token: token to test

        Returns:
            True if this token is an operator, False otherwise
        """

        # Token is an operator
        return token and token.lower() in Token.OPERATORS

    @staticmethod
    def isquoted(token):
        """
        Checks if token is quoted.

        Args:
            token: token to test

        Returns:
            True if this token is quoted, False otherwise
        """

        # Token is quoted
        return token.startswith(("'", '"')) and token.endswith(("'", '"'))

    @staticmethod
    def isseparator(token):
        """
        Checks if token is a separator token.

        Args:
            token to test

        Returns:
            True if this token is a separator, False otherwise
        """

        # Token is a comma
        return token == ","

    @staticmethod
    def issimilar(tokens, x, similar):
        """
        Checks if tokens[x] is a similar() function.

        Args:
            tokens: input tokens
            x: current position
            similar: list where similar function call parameters are stored, can be None in which case similar processing is skipped

        Returns:
            True if tokens[x] is a similar clause
        """

        # True if a "similar" token is followed by an open paren
        return similar is not None and tokens[x].lower() == "similar" and Token.get(tokens, x + 1) == "("

    @staticmethod
    def issortorder(token):
        """
        Checks if token is a sort order token.

        Args:
            token: token to test

        Returns:
            True if this token is a sort order operator, False otherwise
        """

        # Token is a sort order operator
        return token and token.lower() in Token.SORT_ORDER

    @staticmethod
    def normalize(token):
        """
        Applies a normalization algorithm to the input token as follows:
            - Strip single and double quotes
            - Make lowercase

        Args:
            token: input token

        Returns:
            normalized token
        """

        # Lowercase, replace and return
        return token.lower().replace("'", "").replace('"', "")

    @staticmethod
    def wrapspace(text, token):
        """
        Applies whitespace wrapping rules to token.

        Args:
            text: current text buffer
            token: token to add

        Returns:
            token with whitespace rules applied
        """

        # Wildcards have no whitespace. Need special case since * is also multiply which does have whitespace.
        if token in ["*"] and (not text or text.endswith((" ", "("))):
            return token

        # Operator whitespace
        if Token.isoperator(token) or Token.islogicseparator(token) or token.lower() in ["in"]:
            return f" {token} " if not text.endswith(" ") else f"{token} "

        # Comma whitespace
        if Token.isseparator(token):
            return f"{token} "

        # No whitespace if any of the following is True
        if not text or text.endswith((" ", "(", "[")) or token in ["(", "[", ")", "]"] or token.startswith("."):
            return token

        # Default is to add leading whitespace
        return f" {token}"



"""
Defines API paths for objects endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/objects")
def objects(file: str):
    """
    Applies object detection/image classification models to images.

    Args:
        file: file to process

    Returns:
        list of (label, score) elements
    """

    return application.get().pipeline("objects", (file,))


@router.post("/batchobjects")
def batchobjects(files: List[str] = Body(...)):
    """
    Applies object detection/image classification models to images.

    Args:
        files: list of files to process

    Returns:
        list of (label, score) elements
    """

    return application.get().pipeline("objects", (files,))



"""
Vectors module
"""

import json
import os
import tempfile
import uuid

import numpy as np

from ..pipeline import Tokenizer

from .recovery import Recovery


class Vectors:
    """
    Base class for vector models. Vector models transform input content into numeric vectors.
    """

    def __init__(self, config, scoring, models):
        """
        Creates a new vectors instance.

        Args:
            config: vector configuration
            scoring: optional scoring instance for term weighting
            models: models cache
        """

        # Store parameters
        self.config = config
        self.scoring = scoring
        self.models = models

        if config:
            # Detect if this is an initialized configuration
            self.initialized = "dimensions" in config

            # Enables optional string tokenization
            self.tokenize = config.get("tokenize")

            # Load model
            self.model = self.load(config.get("path"))

            # Encode batch size - controls underlying model batch size when encoding vectors
            self.encodebatch = config.get("encodebatch", 32)

            # Embeddings instructions
            self.instructions = config.get("instructions")

            # Truncate embeddings to this dimensionality
            self.dimensionality = config.get("dimensionality")

            # Scalar quantization - supports 1-bit through 8-bit quantization
            quantize = config.get("quantize")
            self.qbits = max(min(quantize, 8), 1) if isinstance(quantize, int) and not isinstance(quantize, bool) else None

    def loadmodel(self, path):
        """
        Loads vector model at path.

        Args:
            path: path to vector model

        Returns:
            vector model
        """

        raise NotImplementedError

    def encode(self, data, category=None):
        """
        Encodes a batch of data using vector model.

        Args:
            data: batch of data
            category: optional category for instruction-based embeddings

        Return:
            transformed data
        """

        raise NotImplementedError

    def load(self, path):
        """
        Loads a model using the current configuration. This method will return previously cached models
        if available.

        Returns:
            model
        """

        # Check if model is cached
        if self.models and path in self.models:
            return self.models[path]

        # Create new model
        model = self.loadmodel(path)

        # Store model in cache
        if self.models is not None and path:
            self.models[path] = model

        return model

    def index(self, documents, batchsize=500, checkpoint=None):
        """
        Converts a list of documents to a temporary file with embeddings arrays. Returns a tuple of document ids,
        number of dimensions and temporary file with embeddings.

        Args:
            documents: list of (id, data, tags)
            batchsize: index batch size
            checkpoint: optional checkpoint directory, enables indexing restart

        Returns:
            (ids, dimensions, batches, stream)
        """

        ids, dimensions, batches, stream = [], None, 0, None

        # Generate recovery config if checkpoint is set
        vectorsid = self.vectorsid() if checkpoint else None
        recovery = Recovery(checkpoint, vectorsid, self.loadembeddings) if checkpoint else None

        # Convert all documents to embedding arrays, stream embeddings to disk to control memory usage
        with self.spool(checkpoint, vectorsid) as output:
            stream = output.name
            batch = []
            for document in documents:
                batch.append(document)

                if len(batch) == batchsize:
                    # Convert batch to embeddings
                    uids, dimensions = self.batch(batch, output, recovery)
                    ids.extend(uids)
                    batches += 1

                    batch = []

            # Final batch
            if batch:
                uids, dimensions = self.batch(batch, output, recovery)
                ids.extend(uids)
                batches += 1

        return (ids, dimensions, batches, stream)

    def vectors(self, documents, batchsize=500, checkpoint=None, buffer=None, dtype=None):
        """
        Bulk encodes documents into vectors using index(). Return the data as a mmap-ed array.

        Args:
            documents: list of (id, data, tags)
            batchsize: index batch size
            checkpoint: optional checkpoint directory, enables indexing restart
            buffer: file path used for memmap buffer
            dtype: dtype for buffer

        Returns:
            (ids, dimensions, embeddings)
        """

        # Consume stream and transform documents to vectors
        ids, dimensions, batches, stream = self.index(documents, batchsize, checkpoint)

        # Check that embeddings are available and load as a memmap
        embeddings = None
        if ids:
            # Write batches
            embeddings = np.memmap(buffer, dtype=dtype, shape=(len(ids), dimensions), mode="w+")
            with open(stream, "rb") as queue:
                x = 0
                for _ in range(batches):
                    batch = self.loadembeddings(queue)
                    embeddings[x : x + batch.shape[0]] = batch
                    x += batch.shape[0]

        # Remove temporary file (if checkpointing is disabled)
        if not checkpoint:
            os.remove(stream)

        return (ids, dimensions, embeddings)

    def close(self):
        """
        Closes this vectors instance.
        """

        self.model = None

    def transform(self, document):
        """
        Transforms document into an embeddings vector.

        Args:
            document: (id, data, tags)

        Returns:
            embeddings vector
        """

        # Prepare input document for vectors model and build embeddings
        return self.batchtransform([document])[0]

    def batchtransform(self, documents, category=None):
        """
        Transforms batch of documents into embeddings vectors.

        Args:
            documents: list of documents used to build embeddings
            category: category for instruction-based embeddings

        Returns:
            embeddings vectors
        """

        # Prepare input documents for vectors model
        documents = [self.prepare(data, category) for _, data, _ in documents]

        # Skip encoding data if it's already an array
        if documents and isinstance(documents[0], np.ndarray):
            return np.array(documents, dtype=np.float32)

        return self.vectorize(documents, category)

    def dot(self, queries, data):
        """
        Calculates the dot product similarity between queries and documents. This method
        assumes each of the inputs are normalized.

        Args:
            queries: queries
            data: search data

        Returns:
            dot product scores
        """

        return np.dot(queries, data.T).tolist()

    def vectorsid(self):
        """
        Generates vectors uid for this vectors instance.

        Returns:
            vectors uid
        """

        # Select config options that determine uniqueness
        select = ["path", "method", "tokenizer", "maxlength", "tokenize", "instructions", "dimensionality", "quantize"]
        config = {k: v for k, v in self.config.items() if k in select}
        config.update(self.config.get("vectors", {}))

        # Generate a deterministic UUID
        return str(uuid.uuid5(uuid.NAMESPACE_DNS, json.dumps(config, sort_keys=True)))

    def spool(self, checkpoint, vectorsid):
        """
        Opens a spool file for queuing generated vectors.

        Args:
            checkpoint: optional checkpoint directory, enables indexing restart
            vectorsid: vectors uid for current configuration

        Returns:
            vectors spool file
        """

        # Spool to vectors checkpoint file
        if checkpoint:
            os.makedirs(checkpoint, exist_ok=True)
            return open(f"{checkpoint}/{vectorsid}", "wb")

        # Spool to temporary file
        return tempfile.NamedTemporaryFile(mode="wb", suffix=".npy", delete=False)

    def batch(self, documents, output, recovery):
        """
        Builds a batch of embeddings.

        Args:
            documents: list of documents used to build embeddings
            output: output temp file to store embeddings
            recovery: optional recovery instance

        Returns:
            (ids, dimensions) list of ids and number of dimensions in embeddings
        """

        # Extract ids and prepare input documents for vectors model
        ids = [uid for uid, _, _ in documents]
        documents = [self.prepare(data, "data") for _, data, _ in documents]
        dimensions = None

        # Attempt to read embeddings from a recovery file
        embeddings = recovery() if recovery else None
        embeddings = self.vectorize(documents, "data") if embeddings is None else embeddings
        if embeddings is not None:
            dimensions = embeddings.shape[1]
            self.saveembeddings(output, embeddings)

        return (ids, dimensions)

    def prepare(self, data, category=None):
        """
        Prepares input data for vector model.

        Args:
            data: input data
            category: category for instruction-based embeddings

        Returns:
            data formatted for vector model
        """

        # Prepares tokens for the model
        data = self.tokens(data)

        # Default instruction category
        category = category if category else "query"

        # Prepend instructions, if applicable
        if self.instructions and category in self.instructions and isinstance(data, str):
            # Prepend category instruction
            data = f"{self.instructions[category]}{data}"

        return data

    def tokens(self, data):
        """
        Prepare data as tokens model can accept.

        Args:
            data: input data

        Returns:
            tokens formatted for model
        """

        # Optional string tokenization
        if self.tokenize and isinstance(data, str):
            data = Tokenizer.tokenize(data)

        # Convert token list to string
        if isinstance(data, list):
            data = " ".join(data)

        return data

    def vectorize(self, data, category=None):
        """
        Runs data vectorization, which consists of the following steps.

          1. Encode data into vectors using underlying model
          2. Truncate vectors, if necessary
          3. Normalize vectors
          4. Quantize vectors, if necessary

        Args:
            data: input data
            category: category for instruction-based embeddings

        Returns:
            embeddings vectors
        """

        # Default instruction category
        category = category if category else "query"

        # Transform data into vectors
        embeddings = self.encode(data, category)

        if embeddings is not None:
            # Truncate embeddings, if necessary
            if self.dimensionality and self.dimensionality < embeddings.shape[1]:
                embeddings = self.truncate(embeddings)

            # Normalize data
            embeddings = self.normalize(embeddings)

            # Apply quantization, if necessary
            if self.qbits:
                embeddings = self.quantize(embeddings)

        return embeddings

    def loadembeddings(self, f):
        """
        Loads embeddings from file.

        Args:
            f: file to load from

        Returns:
            embeddings
        """

        return np.load(f, allow_pickle=False)

    def saveembeddings(self, f, embeddings):
        """
        Saves embeddings to output.

        Args:
            f: output file
            embeddings: embeddings to save
        """

        np.save(f, embeddings, allow_pickle=False)

    def truncate(self, embeddings):
        """
        Truncates embeddings to the configured dimensionality.

        This is only useful for models trained to store more important information in
        earlier dimensions such as Matryoshka Representation Learning (MRL).

        Args:
            embeddings: input embeddings

        Returns:
            truncated embeddings
        """

        return embeddings[:, : self.dimensionality]

    def normalize(self, embeddings):
        """
        Normalizes embeddings using L2 normalization. Operation applied directly on array.

        Args:
            embeddings: input embeddings

        Returns:
            embeddings
        """

        # Calculation is different for matrices vs vectors
        if len(embeddings.shape) > 1:
            embeddings /= np.linalg.norm(embeddings, axis=1)[:, np.newaxis]
        else:
            embeddings /= np.linalg.norm(embeddings)

        return embeddings

    def quantize(self, embeddings):
        """
        Quantizes embeddings using scalar quantization.

        Args:
            embeddings: input embeddings

        Returns:
            quantized embeddings
        """

        # Scale factor is midpoint in range
        factor = 2 ** (self.qbits - 1)

        # Quantize to uint8
        scalars = embeddings * factor
        scalars = scalars.clip(-factor, factor - 1) + factor
        scalars = scalars.astype(np.uint8)

        # Transform uint8 to bits
        bits = np.unpackbits(scalars.reshape(-1, 1), axis=1)

        # Remove unused bits (i.e. for 3-bit quantization, the leading 5 bits are removed)
        bits = bits[:, -self.qbits :]

        # Reshape using original data dimensions and pack bits into uint8 array
        return np.packbits(bits.reshape(embeddings.shape[0], embeddings.shape[1] * self.qbits), axis=1)



"""
Llama module
"""

import os

from huggingface_hub import hf_hub_download

# Conditional import
try:
    import llama_cpp as llama

    LLAMA_CPP = True
except ImportError:
    LLAMA_CPP = False

from .generation import Generation


class LlamaCpp(Generation):
    """
    llama.cpp generative model.
    """

    @staticmethod
    def ismodel(path):
        """
        Checks if path is a llama.cpp model.

        Args:
            path: input path

        Returns:
            True if this is a llama.cpp model, False otherwise
        """

        return isinstance(path, str) and path.lower().endswith(".gguf")

    def __init__(self, path, template=None, **kwargs):
        super().__init__(path, template, **kwargs)

        if not LLAMA_CPP:
            raise ImportError('llama.cpp is not available - install "pipeline" extra to enable')

        # Check if this is a local path, otherwise download from the HF Hub
        path = path if os.path.exists(path) else self.download(path)

        # Create llama.cpp instance
        self.llm = self.create(path, **kwargs)

    def stream(self, texts, maxlength, stream, stop, **kwargs):
        for text in texts:
            yield from (
                self.messages(text, maxlength, stream, stop, **kwargs)
                if isinstance(text, list)
                else self.prompt(text, maxlength, stream, stop, **kwargs)
            )

    def download(self, path):
        """
        Downloads path from the Hugging Face Hub.

        Args:
            path: full model path

        Returns:
            local cached model path
        """

        # Split into parts
        parts = path.split("/")

        # Calculate repo id split
        repo = 2 if len(parts) > 2 else 1

        # Download and cache file
        return hf_hub_download(repo_id="/".join(parts[:repo]), filename="/".join(parts[repo:]))

    def create(self, path, **kwargs):
        """
        Creates a new llama.cpp model instance.

        Args:
            path: path to model
            kwargs: additional keyword args

        Returns:
            llama.cpp instance
        """

        # Default n_ctx=0 if not already set. This sets n_ctx = n_ctx_train.
        kwargs["n_ctx"] = kwargs.get("n_ctx", 0)

        # Default GPU layers if not already set
        kwargs["n_gpu_layers"] = kwargs.get("n_gpu_layers", -1 if kwargs.get("gpu", os.environ.get("LLAMA_NO_METAL") != "1") else 0)

        # Default verbose flag
        kwargs["verbose"] = kwargs.get("verbose", False)

        # Create llama.cpp instance
        try:
            return llama.Llama(model_path=path, **kwargs)
        except ValueError as e:
            # Fallback to default n_ctx when not enough memory for n_ctx = n_ctx_train
            if not kwargs["n_ctx"]:
                kwargs.pop("n_ctx")
                return llama.Llama(model_path=path, **kwargs)

            # Raise exception if n_ctx manually specified
            raise e

    def messages(self, messages, maxlength, stream, stop, **kwargs):
        """
        Processes a list of messages.

        Args:
            messages: list of dictionaries with `role` and `content` key-values
            maxlength: maximum sequence length
            stream: stream response if True, defaults to False
            stop: list of stop strings
            kwargs: additional generation keyword arguments

        Returns:
            generated text
        """

        # LLM call with messages
        result = self.llm.create_chat_completion(messages=messages, max_tokens=maxlength, stream=stream, stop=stop, **kwargs)

        # Stream response
        yield from self.response(result if stream else [result])

    def prompt(self, text, maxlength, stream, stop, **kwargs):
        """
        Processes a prompt.

        Args:
            prompt: prompt text
            maxlength: maximum sequence length
            stream: stream response if True, defaults to False
            stop: list of stop strings
            kwargs: additional generation keyword arguments

        Returns:
            generated text
        """

        # LLM call with prompt
        result = self.llm(text, max_tokens=maxlength, stream=stream, stop=stop, **kwargs)

        # Stream response
        yield from self.response(result if stream else [result])



"""
Execute module
"""

from multiprocessing.pool import Pool, ThreadPool

import torch.multiprocessing


class Execute:
    """
    Supports sequential, multithreading and multiprocessing based execution of tasks.
    """

    def __init__(self, workers=None):
        """
        Creates a new execute instance. Functions can be executed sequentially, in a thread pool
        or in a process pool. Once created, the thread and/or process pool will stay open until the
        close method is called.

        Args:
            workers: number of workers for thread/process pools
        """

        # Number of workers to use in thread/process pools
        self.workers = workers

        self.thread = None
        self.process = None

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, etype, value, traceback):
        self.close()

    def run(self, method, function, args):
        """
        Runs multiple calls of function for each tuple in args. The method parameter controls if the calls are
        sequential (method = None), multithreaded (method = "thread") or with multiprocessing (method="process").

        Args:
            method: run method - "thread" for multithreading, "process" for multiprocessing, otherwise runs sequentially
            function: function to run
            args: list of tuples with arguments to each call
        """

        # Concurrent processing
        if method and len(args) > 1:
            pool = self.pool(method)
            if pool:
                return pool.starmap(function, args, 1)

        # Sequential processing
        return [function(*arg) for arg in args]

    def pool(self, method):
        """
        Gets a handle to a concurrent processing pool. This method will create the pool if it doesn't already exist.

        Args:
            method: pool type - "thread" or "process"

        Returns:
            concurrent processing pool or None if no pool of that type available
        """

        if method == "thread":
            if not self.thread:
                self.thread = ThreadPool(self.workers)

            return self.thread

        if method == "process":
            if not self.process:
                # Importing torch.multiprocessing will register torch shared memory serialization for cuda
                self.process = Pool(self.workers, context=torch.multiprocessing.get_context("spawn"))

            return self.process

        return None

    def close(self):
        """
        Closes concurrent processing pools.
        """

        if hasattr(self, "thread") and self.thread:
            self.thread.close()
            self.thread.join()
            self.thread = None

        if hasattr(self, "process") and self.process:
            self.process.close()
            self.process.join()
            self.process = None



"""
Ids module
"""


class Ids:
    """
    Resolves internal ids for lists of ids.
    """

    def __init__(self, embeddings):
        """
        Create a new ids action.

        Args:
            embeddings: embeddings instance
        """

        self.database = embeddings.database
        self.ids = embeddings.ids

    def __call__(self, ids):
        """
        Resolve internal ids.

        Args:
            ids: ids

        Returns:
            internal ids
        """

        # Resolve ids using database if available, otherwise fallback to embeddings ids
        results = self.database.ids(ids) if self.database else self.scan(ids)

        # Create dict of id: [iids] given there is a one to many relationship
        ids = {}
        for iid, uid in results:
            if uid not in ids:
                ids[uid] = []
            ids[uid].append(iid)

        return ids

    def scan(self, ids):
        """
        Scans embeddings ids array for matches when content is disabled.

        Args:
            ids: search ids

        Returns:
            internal ids
        """

        # Find existing ids
        indices = []
        for uid in ids:
            indices.extend([(index, value) for index, value in enumerate(self.ids) if uid == value])

        return indices



"""
Search module
"""

import logging

from .errors import IndexNotFoundError
from .scan import Scan

# Logging configuration
logger = logging.getLogger(__name__)


class Search:
    """
    Executes a batch search action. A search can be both index and/or database driven.
    """

    def __init__(self, embeddings, indexids=False, indexonly=False):
        """
        Creates a new search action.

        Args:
            embeddings: embeddings instance
            indexids: searches return indexids when True, otherwise run standard search
            indexonly: always runs an index search even when a database is available
        """

        self.embeddings = embeddings
        self.indexids = indexids or indexonly
        self.indexonly = indexonly

        # Alias embeddings attributes
        self.ann = embeddings.ann
        self.batchtransform = embeddings.batchtransform
        self.database = embeddings.database
        self.ids = embeddings.ids
        self.indexes = embeddings.indexes
        self.graph = embeddings.graph
        self.query = embeddings.query
        self.scoring = embeddings.scoring if embeddings.issparse() else None

    def __call__(self, queries, limit=None, weights=None, index=None, parameters=None):
        """
        Executes a batch search for queries. This method will run either an index search or an index + database search
        depending on if a database is available.

        Args:
            queries: list of queries
            limit: maximum results
            weights: hybrid score weights
            index: index name
            parameters: list of dicts of named parameters to bind to placeholders

        Returns:
            list of (id, score) per query for index search
            list of dict per query for an index + database search
            list of graph results for a graph index search
        """

        # Default input parameters
        limit = limit if limit else 3
        weights = weights if weights is not None else 0.5

        # Return empty results if there is no database and indexes
        if not self.ann and not self.scoring and not self.indexes and not self.database:
            return [[]] * len(queries)

        # Default index name if only subindexes set
        if not index and not self.ann and not self.scoring and self.indexes:
            index = self.indexes.default()

        # Graph search
        if self.graph and self.graph.isquery(queries):
            return self.graphsearch(queries, limit, weights, index)

        # Database search
        if not self.indexonly and self.database:
            return self.dbsearch(queries, limit, weights, index, parameters)

        # Default vector index query (sparse, dense or hybrid)
        return self.search(queries, limit, weights, index)

    def search(self, queries, limit, weights, index):
        """
        Executes an index search. When only a sparse index is enabled, this is a a keyword search. When only
        a dense index is enabled, this is an ann search. When both are enabled, this is a hybrid search.

        This method will also query subindexes, if available.

        Args:
            queries: list of queries
            limit: maximum results
            weights: hybrid score weights
            index: index name

        Returns:
            list of (id, score) per query
        """

        # Run against specified subindex
        if index:
            return self.subindex(queries, limit, weights, index)

        # Run against base indexes
        hybrid = self.ann and self.scoring
        dense = self.dense(queries, limit * 10 if hybrid else limit) if self.ann else None
        sparse = self.sparse(queries, limit * 10 if hybrid else limit) if self.scoring else None

        # Combine scores together
        if hybrid:
            # Create weights array if single number passed
            if isinstance(weights, (int, float)):
                weights = [weights, 1 - weights]

            # Create weighted scores
            results = []
            for vectors in zip(dense, sparse):
                uids = {}
                for v, scores in enumerate(vectors):
                    for r, (uid, score) in enumerate(scores if weights[v] > 0 else []):
                        # Initialize score
                        if uid not in uids:
                            uids[uid] = 0.0

                        # Create hybrid score
                        #  - Convex Combination when sparse scores are normalized
                        #  - Reciprocal Rank Fusion (RRF) when sparse scores aren't normalized
                        if self.scoring.isnormalized():
                            uids[uid] += score * weights[v]
                        else:
                            uids[uid] += (1.0 / (r + 1)) * weights[v]

                results.append(sorted(uids.items(), key=lambda x: x[1], reverse=True)[:limit])

            return results

        # Raise an error if when no indexes are available
        if not sparse and not dense:
            raise IndexNotFoundError("No indexes available")

        # Return single query results
        return dense if dense else sparse

    def subindex(self, queries, limit, weights, index):
        """
        Executes a subindex search.

        Args:
            queries: list of queries
            limit: maximum results
            weights: hybrid score weights
            index: index name

        Returns:
            list of (id, score) per query
        """

        # Check that index exists
        if not self.indexes or index not in self.indexes:
            raise IndexNotFoundError(f"Index '{index}' not found")

        # Run subindex search
        results = self.indexes[index].batchsearch(queries, limit, weights)
        return self.resolve(results)

    def dense(self, queries, limit):
        """
        Executes an dense vector search with an approximate nearest neighbor index.

        Args:
            queries: list of queries
            limit: maximum results

        Returns:
            list of (id, score) per query
        """

        # Convert queries to embedding vectors
        embeddings = self.batchtransform((None, query, None) for query in queries)

        # Search approximate nearest neighbor index
        results = self.ann.search(embeddings, limit)

        # Require scores to be greater than 0
        results = [[(i, score) for i, score in r if score > 0] for r in results]

        return self.resolve(results)

    def sparse(self, queries, limit):
        """
        Executes a sparse vector search with a sparse keyword or sparse vector index.

        Args:
            queries: list of queries
            limit: maximum results

        Returns:
            list of (id, score) per query
        """

        # Search sparse index
        results = self.scoring.batchsearch(queries, limit)

        # Require scores to be greater than 0
        results = [[(i, score) for i, score in r if score > 0] for r in results]

        return self.resolve(results)

    def resolve(self, results):
        """
        Resolves index ids. This is only executed when content is disabled.

        Args:
            results: results

        Returns:
            results with resolved ids
        """

        # Map indexids to ids if embeddings ids are available
        if not self.indexids and self.ids:
            return [[(self.ids[i], score) for i, score in r] for r in results]

        return results

    def dbsearch(self, queries, limit, weights, index, parameters):
        """
        Executes an index + database search.

        Args:
            queries: list of queries
            limit: maximum results
            weights: default hybrid score weights
            index: default index name
            parameters: list of dicts of named parameters to bind to placeholders

        Returns:
            list of dict per query
        """

        # Parse queries
        queries = self.parse(queries)

        # Override limit with query limit, if applicable
        limit = max(limit, self.limit(queries))

        # Bulk index scan
        scan = Scan(self.search, limit, weights, index)(queries, parameters)

        # Combine index search results with database search results
        results = []
        for x, query in enumerate(queries):
            # Run the database query, get matching bulk searches for current query
            result = self.database.search(
                query, [r for y, r in scan if x == y], limit, parameters[x] if parameters and parameters[x] else None, self.indexids
            )
            results.append(result)

        return results

    def parse(self, queries):
        """
        Parses a list of database queries.

        Args:
            queries: list of queries

        Returns:
            parsed queries
        """

        # Parsed queries
        parsed = []

        for query in queries:
            # Parse query
            parse = self.database.parse(query)

            # Transform query if SQL not parsed and reparse
            if self.query and "select" not in parse:
                # Generate query
                query = self.query(query)
                logger.debug(query)

                # Reparse query
                parse = self.database.parse(query)

            parsed.append(parse)

        return parsed

    def limit(self, queries):
        """
        Parses the largest LIMIT clause from queries.

        Args:
            queries: list of queries

        Returns:
            largest limit number or 0 if not found
        """

        # Override limit with largest limit from database queries
        qlimit = 0
        for query in queries:
            # Parse out qlimit
            l = query.get("limit")
            if l and l.isdigit():
                l = int(l)

            qlimit = l if l and l > qlimit else qlimit

        return qlimit

    def graphsearch(self, queries, limit, weights, index):
        """
        Executes an index + graph search.

        Args:
            queries: list of queries
            limit: maximum results
            weights: default hybrid score weights
            index: default index name

        Returns:
            graph search results
        """

        # Parse queries
        queries = [self.graph.parse(query) for query in queries]

        # Override limit with query limit, if applicable
        limit = max(limit, self.limit(queries))

        # Bulk index scan
        scan = Scan(self.search, limit, weights, index)(queries, None)

        # Combine index search results with database search results
        for x, query in enumerate(queries):
            # Add search results to query
            query["results"] = [r for y, r in scan if x == y]

        return self.graph.batchsearch(queries, limit, self.indexids)



"""
Terms module
"""

import functools
import os
import sqlite3
import sys

from array import array
from collections import Counter
from threading import RLock

import numpy as np


class Terms:
    """
    Builds, searches and stores memory efficient term frequency sparse arrays for a scoring instance.
    """

    # Term frequency sparse arrays
    CREATE_TERMS = """
        CREATE TABLE IF NOT EXISTS terms (
            term TEXT PRIMARY KEY,
            ids BLOB,
            freqs BLOB
        )
    """

    INSERT_TERM = "INSERT OR REPLACE INTO terms VALUES (?, ?, ?)"
    SELECT_TERMS = "SELECT ids, freqs FROM terms WHERE term = ?"

    # Documents table
    CREATE_DOCUMENTS = """
        CREATE TABLE IF NOT EXISTS documents (
            indexid INTEGER PRIMARY KEY,
            id TEXT,
            deleted INTEGER,
            length INTEGER
        )
    """

    DELETE_DOCUMENTS = "DELETE FROM documents"
    INSERT_DOCUMENT = "INSERT OR REPLACE INTO documents VALUES (?, ?, ?, ?)"
    SELECT_DOCUMENTS = "SELECT indexid, id, deleted, length FROM documents ORDER BY indexid"

    def __init__(self, config, score, idf):
        """
        Creates a new terms index.

        Args:
            config: configuration
            score: score function
            idf: idf weights
        """

        # Terms index configuration
        self.config = config if isinstance(config, dict) else {}
        self.cachelimit = self.config.get("cachelimit", 250000000)
        self.cutoff = self.config.get("cutoff", 0.1)

        # Scoring function
        self.score, self.idf = score, idf

        # Document attributes
        self.ids, self.deletes, self.lengths = [], [], array("q")

        # Terms cache
        self.terms, self.cachesize = {}, 0

        # Terms database
        self.connection, self.cursor, self.path = None, None, None

        # Database thread lock
        self.lock = RLock()

    def insert(self, uid, terms):
        """
        Insert term into index.

        Args:
            uid: document id
            terms: document terms
        """

        # Initialize database, if necessary
        self.initialize()

        # Get next internal index id
        indexid = len(self.ids)

        # Calculate term frequency and document length
        freqs, length = Counter(terms), len(terms)

        # Add document terms
        for term, count in freqs.items():
            # Add term entry
            self.add(indexid, term, count)

            # Each term and freq is a 8-bit signed long long
            self.cachesize += 16

        # Flush cached terms to the database
        if self.cachesize >= self.cachelimit:
            self.index()

        # Save id and length
        self.ids.append(uid)
        self.lengths.append(length)

    def delete(self, ids):
        """
        Mark ids as deleted. This prevents deleted results from showing up in search results.
        The data is not removed from the underlying term frequency sparse arrays.

        Args:
            ids: ids to delete
        """

        # Set index ids as deleted
        self.deletes.extend([self.ids.index(i) for i in ids])

    def index(self):
        """
        Saves any remaining cached terms to the database.
        """

        for term, (nuids, nfreqs) in self.terms.items():
            # Retrieve existing uids/freqs
            uids, freqs = self.lookup(term)

            if uids:
                uids.extend(nuids)
                freqs.extend(nfreqs)
            else:
                uids, freqs = nuids, nfreqs

            # Always save as little endian
            if sys.byteorder == "big":
                uids.byteswap()
                freqs.byteswap()

            # Insert or replace term
            self.cursor.execute(Terms.INSERT_TERM, [term, uids.tobytes(), freqs.tobytes()])

        # Clear cached weights
        self.weights.cache_clear()

        # Reset term cache size
        self.terms, self.cachesize = {}, 0

    def search(self, terms, limit):
        """
        Searches term index a term-at-a-time. Each term frequency sparse array is retrieved
        and used to calculate term match scores.

        This method calculates term scores in two steps as shown below.

          1. Query and score less common term scores first
          2. Merge in common term scores for all documents matching the first query

        This is similar to the common terms query in Apache Lucene.

        Args:
            terms: query terms
            limit: maximum results

        Returns:
            list of (id, score)
        """

        # Initialize scores array
        scores = np.zeros(len(self.ids), dtype=np.float32)

        # Score less common terms
        terms, skipped, hasscores = Counter(terms), {}, False
        for term, freq in terms.items():
            # Compute or lookup term weights
            uids, weights = self.weights(term)
            if uids is not None:
                # Term considered common if it appears in more than 10% of index
                if len(uids) <= self.cutoff * len(self.ids):
                    # Add scores
                    scores[uids] += freq * weights

                    # Set flag that scores have been calculated for at least one term
                    hasscores = True
                else:
                    skipped[term] = freq

        # Merge in common term scores and return top n matches
        return self.topn(scores, limit, hasscores, skipped)

    def count(self):
        """
        Number of elements in the scoring index.

        Returns:
            count
        """

        return len(self.ids) - len(self.deletes)

    def load(self, path):
        """
        Loads terms database from path. This method loads document attributes into memory.

        Args:
            path: path to read terms database
        """

        # Load an existing terms database
        self.connection = self.connect(path)
        self.cursor = self.connection.cursor()
        self.path = path

        # Load document attributes
        self.ids, self.deletes, self.lengths = [], [], array("q")

        self.cursor.execute(Terms.SELECT_DOCUMENTS)
        for indexid, uid, deleted, length in self.cursor:
            # Index id - id
            self.ids.append(uid)

            # Deleted flag
            if deleted:
                self.deletes.append(indexid)

            # Index id - length
            self.lengths.append(length)

        # Cast ids to int if every id is an integer
        if all(uid.isdigit() for uid in self.ids):
            self.ids = [int(uid) for uid in self.ids]

        # Clear cache
        self.weights.cache_clear()

    def save(self, path):
        """
        Saves terms database to path. This method creates or replaces document attributes into the database.

        Args:
            path: path to write terms database
        """

        # Clear documents table
        self.cursor.execute(Terms.DELETE_DOCUMENTS)

        # Save document attributes
        for i, uid in enumerate(self.ids):
            self.cursor.execute(Terms.INSERT_DOCUMENT, [i, uid, 1 if i in self.deletes else 0, self.lengths[i]])

        # Temporary database
        if not self.path:
            # Save temporary database
            self.connection.commit()

            # Copy data from current to new
            connection = self.copy(path)

            # Close temporary database
            self.connection.close()

            # Point connection to new connection
            self.connection = connection
            self.cursor = self.connection.cursor()
            self.path = path

        # Paths are equal, commit changes
        elif self.path == path:
            self.connection.commit()

        # New path is different from current path, copy data and continue using current connection
        else:
            self.copy(path).close()

    def close(self):
        """
        Close and free resources used by this instance.
        """

        # Close connection
        if self.connection:
            self.connection.close()

    def initialize(self):
        """
        Creates connection and initial database schema if no connection exists.
        """

        if not self.connection:
            # Create term database
            self.connection = self.connect()
            self.cursor = self.connection.cursor()

            # Create initial schema
            self.cursor.execute(Terms.CREATE_TERMS)
            self.cursor.execute(Terms.CREATE_DOCUMENTS)

    def connect(self, path=""):
        """
        Creates a new term database connection.

        Args:
            path: path to term database file

        Returns:
            connection
        """

        connection = sqlite3.connect(path, check_same_thread=False)

        # Enable WAL mode, if necessary
        if self.config.get("wal"):
            connection.execute("PRAGMA journal_mode=WAL")

        return connection

    def copy(self, path):
        """
        Copies content from current terms database into target.

        Args:
            path: target database path

        Returns:
            new database connection
        """

        # Delete existing file, if necessary
        if os.path.exists(path):
            os.remove(path)

        # Create new connection
        connection = self.connect(path)

        if self.connection.in_transaction:
            # The backup call will hang if there are uncommitted changes, need to copy over
            # with iterdump (which is much slower)
            for sql in self.connection.iterdump():
                connection.execute(sql)
        else:
            # Database is up to date, can do a more efficient copy with SQLite C API
            self.connection.backup(connection)

        return connection

    def add(self, indexid, term, freq):
        """
        Adds a term frequency entry.

        Args:
            indexid: internal index id
            term: term
            freq: term frequency
        """

        # Get or create uids and freqs arrays
        if term not in self.terms:
            self.terms[term] = (array("q"), array("q"))

        # Append uids and freqs
        ids, freqs = self.terms[term]
        ids.append(indexid)
        freqs.append(freq)

    def lookup(self, term):
        """
        Retrieves a term frequency sparse array.

        Args:
            term: term to lookup

        Returns:
            term frequency sparse array
        """

        uids, freqs = None, None

        result = self.cursor.execute(Terms.SELECT_TERMS, [term]).fetchone()
        if result:
            uids, freqs = (array("q"), array("q"))
            uids.frombytes(result[0])
            freqs.frombytes(result[1])

            # Storage format is always little endian
            if sys.byteorder == "big":
                uids.byteswap()
                freqs.byteswap()

        return uids, freqs

    @functools.lru_cache(maxsize=500)
    def weights(self, term):
        """
        Computes a term weights sparse array for term. This method is wrapped with a least recently used cache,
        which will return common term weights from the cache.

        Args:
            term: term

        Returns:
            term weights sparse array
        """

        lengths = np.frombuffer(self.lengths, dtype=np.int64)

        with self.lock:
            uids, freqs = self.lookup(term)
            weights = None

        if uids:
            uids = np.frombuffer(uids, dtype=np.int64)
            weights = self.score(np.frombuffer(freqs, dtype=np.int64), self.idf[term], lengths[uids]).astype(np.float32)

        return uids, weights

    def topn(self, scores, limit, hasscores, skipped):
        """
        Get topn scores from an partial scores array.

        Args:
            scores: partial scores array with scores for less common terms
            limit: maximum results
            hasscores: True if partial scores array has any nonzero scores, False otherwise
            skipped: terms skipped in initial query

        Returns:
            topn scores
        """

        # Calculate topn candidates to consider
        # Require at least one positive score, set topn to smaller of limit * 5 or number of scores
        topn = min(len(scores), limit * 5)

        # Get topn candidates, allows for score shifting when adding in common term scores
        matches = self.candidates(scores, topn)

        # Merge in scores for more common terms
        self.merge(scores, matches, hasscores, skipped)

        # Get topn candidates since it was initially skipped above
        if not hasscores:
            matches = self.candidates(scores, topn)

        # Reorder matches using updated scores
        matches = matches[np.argsort(-scores[matches])]

        # Combine ids with scores. Require score > 0.
        return [(self.ids[x], float(scores[x])) for x in matches[:limit] if scores[x] > 0]

    def merge(self, scores, matches, hasscores, terms):
        """
        Merges common term scores into scores array.

        Args:
            scores: partial scores array
            matches: current matches, if any
            hasscores: True if scores has current matches, False otherwise
            terms: common terms
        """

        for term, freq in terms.items():
            # Compute or lookup term weights
            uids, weights = self.weights(term)

            # Filter to topn matches when partial scores array has nonzero scores
            if hasscores:
                # Find indices in match ids for uids
                indices = np.searchsorted(uids, matches)

                # Filter matches that don't exist in uids
                indices = [x for i, x in enumerate(indices) if x < len(uids) and uids[x] == matches[i]]

                # Filter to matching uids and weights
                uids, weights = uids[indices], weights[indices]

            # Update scores
            scores[uids] += freq * weights

    def candidates(self, scores, topn):
        """
        Gets the topn scored candidates. This method ignores deleted documents.

        Args:
            scores: scores array
            topn: topn elements

        Returns:
            topn scored candidates
        """

        # Clear deletes
        scores[self.deletes] = 0

        # Get topn candidates
        return np.argpartition(scores, -topn)[-topn:]



"""
LiteLLM module
"""

import numpy as np

from transformers.utils import cached_file

# Conditional import
try:
    import litellm as api

    LITELLM = True
except ImportError:
    LITELLM = False

from ..base import Vectors


class LiteLLM(Vectors):
    """
    Builds vectors using an external embeddings API via LiteLLM.
    """

    @staticmethod
    def ismodel(path):
        """
        Checks if path is a LiteLLM model.

        Args:
            path: input path

        Returns:
            True if this is a LiteLLM model, False otherwise
        """

        # pylint: disable=W0702
        if isinstance(path, str) and LITELLM:
            debug = api.suppress_debug_info
            try:
                # Suppress debug messages for this test
                api.suppress_debug_info = True
                return api.get_llm_provider(path) and not LiteLLM.ishub(path)
            except:
                return False
            finally:
                # Restore debug info value to original value
                api.suppress_debug_info = debug

        return False

    @staticmethod
    def ishub(path):
        """
        Checks if path is available on the HF Hub.

        Args:
            input path

        Returns:
            True if this is a model on the HF Hub
        """

        # pylint: disable=W0702
        try:
            return cached_file(path_or_repo_id=path, filename="config.json") is not None if "/" in path else False
        except:
            return False

    def __init__(self, config, scoring, models):
        # Check before parent constructor since it calls loadmodel
        if not LITELLM:
            raise ImportError('LiteLLM is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def loadmodel(self, path):
        return None

    def encode(self, data, category=None):
        # Call external embeddings API using LiteLLM
        # Batching is handled server-side
        response = api.embedding(model=self.config.get("path"), input=data, **self.config.get("vectors", {}))

        # Read response into a NumPy array
        return np.array([x["embedding"] for x in response.data], dtype=np.float32)



"""
Model2Vec module
"""

import json

from huggingface_hub.errors import HFValidationError
from transformers.utils import cached_file

# Conditional import
try:
    from model2vec import StaticModel

    MODEL2VEC = True
except ImportError:
    MODEL2VEC = False

from ..base import Vectors


class Model2Vec(Vectors):
    """
    Builds vectors using Model2Vec.
    """

    @staticmethod
    def ismodel(path):
        """
        Checks if path is a Model2Vec model.

        Args:
            path: input path

        Returns:
            True if this is a Model2Vec model, False otherwise
        """

        try:
            # Download file and parse JSON
            path = cached_file(path_or_repo_id=path, filename="config.json")
            if path:
                with open(path, encoding="utf-8") as f:
                    config = json.load(f)
                    return config.get("model_type") == "model2vec"

        # Ignore this error - invalid repo or directory
        except (HFValidationError, OSError):
            pass

        return False

    def __init__(self, config, scoring, models):
        # Check before parent constructor since it calls loadmodel
        if not MODEL2VEC:
            raise ImportError('Model2Vec is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def loadmodel(self, path):
        return StaticModel.from_pretrained(path)

    def encode(self, data, category=None):
        # Additional model arguments
        modelargs = self.config.get("vectors", {})

        # Encode data
        return self.model.encode(data, batch_size=self.encodebatch, **modelargs)



"""
Defines API paths for translation endpoints.
"""

from typing import List, Optional

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/translate")
def translate(text: str, target: Optional[str] = "en", source: Optional[str] = None):
    """
    Translates text from source language into target language.

    Args:
        text: text to translate
        target: target language code, defaults to "en"
        source: source language code, detects language if not provided

    Returns:
        translated text
    """

    return application.get().pipeline("translation", (text, target, source))


@router.post("/batchtranslate")
def batchtranslate(texts: List[str] = Body(...), target: Optional[str] = Body(default="en"), source: Optional[str] = Body(default=None)):
    """
    Translates text from source language into target language.

    Args:
        texts: list of text to translate
        target: target language code, defaults to "en"
        source: source language code, detects language if not provided

    Returns:
        list of translated text
    """

    return application.get().pipeline("translation", (texts, target, source))



"""
Model module
"""

import re

from enum import Enum

from smolagents import ChatMessage, Model, get_clean_message_list, tool_role_conversions
from smolagents.models import get_tool_call_from_text, remove_stop_sequences

from ..pipeline import LLM


class PipelineModel(Model):
    """
    Model backed by a LLM pipeline.
    """

    def __init__(self, path=None, method=None, **kwargs):
        """
        Creates a new LLM model.

        Args:
            path: model path or instance
            method: llm model framework, infers from path if not provided
            kwargs: model keyword arguments
        """

        self.llm = path if isinstance(path, LLM) else LLM(path, method, **kwargs)
        self.maxlength = 8192

        # Set base class parameters
        self.model_id = self.llm.generator.path

        # Call parent constructor
        super().__init__(flatten_messages_as_text=not self.llm.isvision(), **kwargs)

    # pylint: disable=W0613
    def generate(self, messages, stop_sequences=None, response_format=None, tools_to_call_from=None, **kwargs):
        """
        Runs LLM inference. This method signature must match the smolagents specification.

        Args:
            messages: list of messages to run
            stop_sequences: optional list of stop sequences
            response_format: response format to use in the model's response.
            tools_to_call_from: list of tools that the model can use to generate responses.
            kwargs: additional keyword arguments

        Returns:
            result
        """

        # Get clean message list
        messages = self.clean(messages)

        # Get LLM output
        response = self.llm(messages, maxlength=self.maxlength, stop=stop_sequences, **kwargs)

        # Remove stop sequences from LLM output
        if stop_sequences is not None:
            response = remove_stop_sequences(response, stop_sequences)

        # Load response into a chat message
        message = ChatMessage(role="assistant", content=response)

        # Extract first tool action, if necessary
        if tools_to_call_from:
            message.tool_calls = [
                get_tool_call_from_text(
                    re.sub(r".*?Action:(.*?\n\}).*", r"\1", response, flags=re.DOTALL), self.tool_name_key, self.tool_arguments_key
                )
            ]

        return message

    def parameters(self, maxlength):
        """
        Set LLM inference parameters.

        Args:
            maxlength: maximum sequence length
        """

        self.maxlength = maxlength

    def clean(self, messages):
        """
        Gets a clean message list.

        Args:
            messages: input messages

        Returns:
            clean messages
        """

        # Get clean message list
        messages = get_clean_message_list(messages, role_conversions=tool_role_conversions, flatten_messages_as_text=self.flatten_messages_as_text)

        # Ensure all roles are strings and not enums for compability across LLM frameworks
        for message in messages:
            if "role" in message:
                message["role"] = message["role"].value if isinstance(message["role"], Enum) else message["role"]

        return messages



"""
ImageHash module
"""

import numpy as np

# Conditional import
try:
    from PIL import Image
    import imagehash

    PIL = True
except ImportError:
    PIL = False

from ..base import Pipeline


class ImageHash(Pipeline):
    """
    Generates perceptual image hashes. These hashes can be used to detect near-duplicate images. This method is not
    backed by machine learning models and not intended to find conceptually similar images.
    """

    def __init__(self, algorithm="average", size=8, strings=True):
        """
        Creates an ImageHash pipeline.

        Args:
            algorithm: image hashing algorithm (average, perceptual, difference, wavelet, color)
            size: hash size
            strings: outputs hex strings if True (default), otherwise the pipeline returns numpy arrays
        """

        if not PIL:
            raise ImportError('ImageHash pipeline is not available - install "pipeline" extra to enable')

        self.algorithm = algorithm
        self.size = size
        self.strings = strings

    def __call__(self, images):
        """
        Generates perceptual image hashes.

        Args:
            images: image|list

        Returns:
            list of hashes
        """

        # Convert single element to list
        values = [images] if not isinstance(images, list) else images

        # Open images if file strings
        values = [Image.open(image) if isinstance(image, str) else image for image in values]

        # Convert images to hashes
        hashes = [self.ihash(image) for image in values]

        # Return single element if single element passed in
        return hashes[0] if not isinstance(images, list) else hashes

    def ihash(self, image):
        """
        Gets an image hash for image.

        Args:
            image: PIL image

        Returns:
            hash as hex string
        """

        # Apply hash algorithm
        if self.algorithm == "perceptual":
            data = imagehash.phash(image, self.size)
        elif self.algorithm == "difference":
            data = imagehash.dhash(image, self.size)
        elif self.algorithm == "wavelet":
            data = imagehash.whash(image, self.size)
        elif self.algorithm == "color":
            data = imagehash.colorhash(image, self.size)
        else:
            # Default to average hash
            data = imagehash.average_hash(image, self.size)

        # Convert to output data type
        return str(data) if self.strings else data.hash.astype(np.float32).reshape(-1)



"""
Defines API paths for labels endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.post("/label")
def label(text: str = Body(...), labels: List[str] = Body(...)):
    """
    Applies a zero shot classifier to text using a list of labels. Returns a list of
    {id: value, score: value} sorted by highest score, where id is the index in labels.

    Args:
        text: input text
        labels: list of labels

    Returns:
        list of {id: value, score: value} per text element
    """

    return application.get().label(text, labels)


@router.post("/batchlabel")
def batchlabel(texts: List[str] = Body(...), labels: List[str] = Body(...)):
    """
    Applies a zero shot classifier to list of text using a list of labels. Returns a list of
    {id: value, score: value} sorted by highest score, where id is the index in labels per
    text element.

    Args:
        texts: list of text
        labels: list of labels

    Returns:
        list of {id: value score: value} per text element
    """

    return application.get().label(texts, labels)



"""
Pickle module
"""

import os
import logging
import pickle
import warnings

from .base import Serialize

# Logging configuration
logger = logging.getLogger(__name__)


class Pickle(Serialize):
    """
    Pickle serialization.
    """

    def __init__(self, allowpickle=False):
        """
        Creates a new instance for Pickle serialization.

        This class ensures the allowpickle parameter or the `ALLOW_PICKLE` environment variable is True. All methods will
        raise errors if this isn't the case.

        Pickle serialization is OK for local data but it isn't recommended when sharing data externally.

        Args:
            allowpickle: default pickle allow mode, only True with methods that generate local temporary data
        """

        # Parent constructor
        super().__init__()

        # Default allow pickle mode
        self.allowpickle = allowpickle

        # Current pickle protocol
        self.version = 4

    def load(self, path):
        # Load pickled data from path, if allowed
        return super().load(path) if self.allow(path) else None

    def save(self, data, path):
        # Save pickled data to path, if allowed
        if self.allow():
            super().save(data, path)

    def loadstream(self, stream):
        # Load pickled data from stream, if allowed
        return pickle.load(stream) if self.allow() else None

    def savestream(self, data, stream):
        # Save pickled data to stream, if allowed
        if self.allow():
            pickle.dump(data, stream, protocol=self.version)

    def loadbytes(self, data):
        # Load pickled data from bytes, if allowed
        return pickle.loads(data) if self.allow() else None

    def savebytes(self, data):
        # Save pickled data to stream, if allowed
        return pickle.dumps(data, protocol=self.version) if self.allow() else None

    def allow(self, path=None):
        """
        Checks if loading and saving pickled data is allowed. Raises an error if it's not allowed.

        Args:
            path: optional path to add to generated error messages
        """

        enablepickle = self.allowpickle or os.environ.get("ALLOW_PICKLE", "False") in ("True", "1")
        if not enablepickle:
            raise ValueError(
                (
                    "Loading of pickled index data is disabled. "
                    f"`{path if path else 'stream'}` was not loaded. "
                    "Set the env variable `ALLOW_PICKLE=True` to enable loading pickled index data. "
                    "This should only be done for trusted and/or local data."
                )
            )

        if not self.allowpickle:
            warnings.warn(
                (
                    "Loading of pickled data enabled through `ALLOW_PICKLE=True` env variable. "
                    "This setting should only be used with trusted and/or local data. "
                    "Saving this index will replace pickled index data formats with the latest index formats and remove this warning."
                ),
                RuntimeWarning,
            )

        return enablepickle



"""
Factory module
"""

import inspect

from types import FunctionType, MethodType

import mcpadapt.core

from mcpadapt.smolagents_adapter import SmolAgentsAdapter
from smolagents import PythonInterpreterTool, Tool, tool as CreateTool, VisitWebpageTool, WebSearchTool
from transformers.utils import chat_template_utils, TypeHintParsingException

from ...embeddings import Embeddings
from .embeddings import EmbeddingsTool
from .function import FunctionTool


class ToolFactory:
    """
    Methods to create tools.
    """

    # Default toolkit
    DEFAULTS = {"python": PythonInterpreterTool(), "websearch": WebSearchTool(), "webview": VisitWebpageTool()}

    @staticmethod
    def create(config):
        """
        Creates a new list of tools. This method iterates of the `tools` configuration option and creates a Tool instance
        for each entry. This supports the following:

          - Tool instance
          - Dictionary with `name`, `description`, `inputs`, `output` and `target` function configuration
          - String with a tool alias name

        Returns:
            list of tools
        """

        tools = []
        for tool in config.pop("tools", []):
            # Create tool from function and it's documentation
            if not isinstance(tool, Tool) and (isinstance(tool, (FunctionType, MethodType)) or hasattr(tool, "__call__")):
                tool = ToolFactory.createtool(tool)

            # Create tool from input dictionary
            elif isinstance(tool, dict):
                # Get target function
                target = tool.get("target")

                # Create tool from input dictionary
                tool = (
                    EmbeddingsTool(tool)
                    if isinstance(target, Embeddings) or any(x in tool for x in ["container", "path"])
                    else ToolFactory.createtool(target, tool)
                )

            # Get default tool, if applicable
            elif isinstance(tool, str) and tool in ToolFactory.DEFAULTS:
                tool = ToolFactory.DEFAULTS[tool]

            # Support importing MCP tool collections
            elif isinstance(tool, str) and tool.startswith("http"):
                tools.extend(mcpadapt.core.MCPAdapt({"url": tool}, SmolAgentsAdapter()).tools())
                tool = None

            # Add tool
            if tool:
                tools.append(tool)

        return tools

    @staticmethod
    def createtool(target, config=None):
        """
        Creates a new Tool.

        Args:
            target: target object or function
            config: optional tool configuration

        Returns:
            Tool
        """

        try:
            # Try to create using CreateTool function - this fails when no annotations are available
            return CreateTool(target)
        except (TypeHintParsingException, TypeError):
            return ToolFactory.fromdocs(target, config if config else {})

    @staticmethod
    def fromdocs(target, config):
        """
        Creates a tool from method documentation.

        Args:
            target: target object or function
            config: tool configuration

        Returns:
            Tool
        """

        # Get function name and target - use target if it's a function or method, else use target.__call__
        name = target.__name__ if isinstance(target, (FunctionType, MethodType)) or not hasattr(target, "__call__") else target.__class__.__name__
        target = target if isinstance(target, (FunctionType, MethodType)) or not hasattr(target, "__call__") else target.__call__

        # Extract target documentation
        doc = inspect.getdoc(target)
        description, parameters, _ = chat_template_utils.parse_google_format_docstring(doc.strip()) if doc else (None, {}, None)

        # Get list of required parameters
        signature = inspect.signature(target)
        inputs = {}
        for pname, param in signature.parameters.items():
            if param.default == inspect.Parameter.empty and pname in parameters:
                inputs[pname] = {"type": "any", "description": parameters[pname]}

        # Create function tool
        return FunctionTool(
            {
                "name": config.get("name", name.lower()),
                "description": config.get("description", description),
                "inputs": config.get("inputs", inputs),
                "target": config.get("target", target),
            }
        )



"""
Terms module
"""


class Terms:
    """
    Reduces a query statement down to keyword terms. This method extracts the query text from similar clauses if it's a SQL statement.
    Otherwise, the original query is returned.
    """

    def __init__(self, embeddings):
        """
        Create a new terms action.

        Args:
            embeddings: embeddings instance
        """

        self.database = embeddings.database

    def __call__(self, queries):
        """
        Extracts keyword terms from a list of queries.

        Args:
            queries: list of queries

        Returns:
            list of queries reduced down to keyword term strings
        """

        # Parse queries and extract keyword terms for each query
        if self.database:
            terms = []
            for query in queries:
                # Parse query
                parse = self.database.parse(query)

                # Join terms from similar clauses
                terms.append(" ".join(" ".join(s) for s in parse["similar"]))

            return terms

        # Return original query when database is None
        return queries



"""
Llama module
"""

import os

import numpy as np

from huggingface_hub import hf_hub_download

# Conditional import
try:
    from llama_cpp import Llama

    LLAMA_CPP = True
except ImportError:
    LLAMA_CPP = False

from ..base import Vectors


class LlamaCpp(Vectors):
    """
    Builds vectors using llama.cpp.
    """

    @staticmethod
    def ismodel(path):
        """
        Checks if path is a llama.cpp model.

        Args:
            path: input path

        Returns:
            True if this is a llama.cpp model, False otherwise
        """

        return isinstance(path, str) and path.lower().endswith(".gguf")

    def __init__(self, config, scoring, models):
        # Check before parent constructor since it calls loadmodel
        if not LLAMA_CPP:
            raise ImportError('llama.cpp is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def loadmodel(self, path):
        # Check if this is a local path, otherwise download from the HF Hub
        path = path if os.path.exists(path) else self.download(path)

        # Additional model arguments
        modelargs = self.config.get("vectors", {})

        # Default GPU layers if not already set
        modelargs["n_gpu_layers"] = modelargs.get("n_gpu_layers", -1 if self.config.get("gpu", os.environ.get("LLAMA_NO_METAL") != "1") else 0)

        # Create llama.cpp instance
        return Llama(path, n_ctx=0, verbose=modelargs.pop("verbose", False), embedding=True, **modelargs)

    def encode(self, data, category=None):
        # Generate embeddings and return as a NumPy array
        # llama-cpp-python has it's own batching built-in using n_batch parameter
        return np.array(self.model.embed(data), dtype=np.float32)

    def download(self, path):
        """
        Downloads path from the Hugging Face Hub.

        Args:
            path: full model path

        Returns:
            local cached model path
        """

        # Split into parts
        parts = path.split("/")

        # Calculate repo id split
        repo = 2 if len(parts) > 2 else 1

        # Download and cache file
        return hf_hub_download(repo_id="/".join(parts[:repo]), filename="/".join(parts[repo:]))



"""
TextToAudio module
"""

from ..hfpipeline import HFPipeline
from .signal import Signal, SCIPY


class TextToAudio(HFPipeline):
    """
    Generates audio from text.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, rate=None, **kwargs):
        if not SCIPY:
            raise ImportError('TextToAudio pipeline is not available - install "pipeline" extra to enable.')

        # Call parent constructor
        super().__init__("text-to-audio", path, quantize, gpu, model, **kwargs)

        # Target sample rate, defaults to model sample rate
        self.rate = rate

    def __call__(self, text, maxlength=512):
        """
        Generates audio from text.

        This method supports text as a string or a list. If the input is a string,
        the return type is a single audio output. If text is a list, the return type is a list.

        Args:
            text: text|list
            maxlength: maximum audio length to generate

        Returns:
            list of (audio, sample rate)
        """

        # Format inputs
        texts = [text] if isinstance(text, str) else text

        # Run pipeline
        results = [self.convert(x) for x in self.pipeline(texts, forward_params={"max_new_tokens": maxlength})]

        # Extract results
        return results[0] if isinstance(text, str) else results

    def convert(self, result):
        """
        Converts audio result to target sample rate for this pipeline, if set.

        Args:
            result: dict with audio samples and sample rate

        Returns:
            (audio, sample rate)
        """

        audio, rate = result["audio"].squeeze(), result["sampling_rate"]
        return (Signal.resample(audio, rate, self.rate), self.rate) if self.rate else (audio, rate)



"""
ImageTask module
"""

import re

# Conditional import
try:
    from PIL import Image

    PIL = True
except ImportError:
    PIL = False

from .file import FileTask


class ImageTask(FileTask):
    """
    Task that processes image file urls
    """

    def register(self):
        """
        Checks if required dependencies are installed.
        """

        if not PIL:
            raise ImportError('ImageTask is not available - install "workflow" extra to enable')

    def accept(self, element):
        # Only accept image files
        return super().accept(element) and re.search(r"\.(gif|bmp|jpg|jpeg|png|webp)$", element.lower())

    def prepare(self, element):
        return Image.open(super().prepare(element))



"""
Defines API paths for agent endpoints.
"""

from typing import Optional

from fastapi import APIRouter, Body
from fastapi.responses import StreamingResponse

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.post("/agent")
def agent(name: str = Body(...), text: str = Body(...), maxlength: Optional[int] = Body(default=None), stream: Optional[bool] = Body(default=None)):
    """
    Executes a named agent for input text.

    Args:
        name: agent name
        text: instructions to run
        maxlength: maximum sequence length
        stream: stream response if True, defaults to False

    Returns:
        response text
    """

    # Build keyword arguments
    kwargs = {key: value for key, value in [("stream", stream), ("maxlength", maxlength)] if value}

    # Run agent
    result = application.get().agent(name, text, **kwargs)

    # Handle both standard and streaming responses
    return StreamingResponse(result) if stream else result



"""
Encoder factory module
"""

from ...util import Resolver

from .base import Encoder
from .serialize import SerializeEncoder


class EncoderFactory:
    """
    Encoder factory. Creates new Encoder instances.
    """

    @staticmethod
    def get(encoder):
        """
        Gets a new instance of encoder class.

        Args:
            encoder: Encoder instance class

        Returns:
            Encoder class
        """

        # Local task if no package
        if "." not in encoder:
            # Get parent package
            encoder = ".".join(__name__.split(".")[:-1]) + "." + encoder.capitalize() + "Encoder"

        return Resolver()(encoder)

    @staticmethod
    def create(encoder):
        """
        Creates a new Encoder instance.

        Args:
            encoder: Encoder instance class

        Returns:
            Encoder
        """

        # Return default encoder
        if encoder is True:
            return Encoder()

        # Supported serialization methods
        if encoder in ["messagepack", "pickle"]:
            return SerializeEncoder(encoder)

        # Get Encoder instance
        return EncoderFactory.get(encoder)()



"""
Query module
"""

from transformers import AutoTokenizer, AutoModelForSeq2SeqLM, T5ForConditionalGeneration


class Query:
    """
    Query translation model.
    """

    def __init__(self, path, prefix=None, maxlength=512):
        """
        Creates a query translation model.

        Args:
            path: path to query model
            prefix: text prefix
            maxlength: max sequence length to generate
        """

        self.tokenizer = AutoTokenizer.from_pretrained(path)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(path)

        # Default prefix if not provided for T5 models
        if not prefix and isinstance(self.model, T5ForConditionalGeneration):
            prefix = "translate English to SQL: "

        self.prefix = prefix
        self.maxlength = maxlength

    def __call__(self, query):
        """
        Runs query translation model.

        Args:
            query: input query

        Returns:
            transformed query
        """

        # Add prefix, if necessary
        if self.prefix:
            query = f"{self.prefix}{query}"

        # Tokenize and generate text using model
        features = self.tokenizer([query], return_tensors="pt")
        output = self.model.generate(input_ids=features["input_ids"], attention_mask=features["attention_mask"], max_length=self.maxlength)

        # Decode tokens to text
        result = self.tokenizer.decode(output[0], skip_special_tokens=True)

        # Clean and return generated text
        return self.clean(result)

    def clean(self, text):
        """
        Applies a series of rules to clean generated text.

        Args:
            text: input text

        Returns:
            clean text
        """

        return text.replace("$=", "<=")



"""
LLM module
"""

import logging

from .factory import GenerationFactory

from ..base import Pipeline

# Logging configuration
logger = logging.getLogger(__name__)


class LLM(Pipeline):
    """
    Pipeline for running large language models (LLMs). This class supports the following LLM backends:

      - Local LLMs with Hugging Face Transformers
      - Local LLMs with llama.cpp
      - Remote API LLMs with LiteLLM
      - Custom generation implementations
    """

    def __init__(self, path=None, method=None, **kwargs):
        """
        Creates a new LLM.

        Args:
            path: model path
            method: llm model framework, infers from path if not provided
            kwargs: model keyword arguments
        """

        # Default LLM if not provided
        path = path if path else "google/flan-t5-base"

        # Generation instance
        self.generator = GenerationFactory.create(path, method, **kwargs)

    def __call__(self, text, maxlength=512, stream=False, stop=None, defaultrole="prompt", stripthink=False, **kwargs):
        """
        Generates text. Supports the following input formats:

          - String or list of strings (instruction-tuned models must follow chat templates)
          - List of dictionaries with `role` and `content` key-values or lists of lists

        Args:
            text: text|list
            maxlength: maximum sequence length
            stream: stream response if True, defaults to False
            stop: list of stop strings, defaults to None
            defaultrole: default role to apply to text inputs (prompt for raw prompts (default) or user for user chat messages)
            stripthink: strip thinking tags, defaults to False
            kwargs: additional generation keyword arguments

        Returns:
            generated text
        """

        # Debug logging
        logger.debug(text)

        # Run LLM generation
        return self.generator(text, maxlength, stream, stop, defaultrole, stripthink, **kwargs)

    def isvision(self):
        """
        Returns True if this LLM supports vision operations.

        Returns:
            True if this is a vision model
        """

        return self.generator.isvision()



"""
HNSW module
"""

import numpy as np

# Conditional import
try:
    # pylint: disable=E0611
    from hnswlib import Index

    HNSWLIB = True
except ImportError:
    HNSWLIB = False

from ..base import ANN


class HNSW(ANN):
    """
    Builds an ANN index using the hnswlib library.
    """

    def __init__(self, config):
        super().__init__(config)

        if not HNSWLIB:
            raise ImportError('HNSW is not available - install "ann" extra to enable')

    def load(self, path):
        # Load index
        self.backend = Index(dim=self.config["dimensions"], space=self.config["metric"])
        self.backend.load_index(path)

    def index(self, embeddings):
        # Inner product is equal to cosine similarity on normalized vectors
        self.config["metric"] = "ip"

        # Lookup index settings
        efconstruction = self.setting("efconstruction", 200)
        m = self.setting("m", 16)
        seed = self.setting("randomseed", 100)

        # Create index
        self.backend = Index(dim=self.config["dimensions"], space=self.config["metric"])
        self.backend.init_index(max_elements=embeddings.shape[0], ef_construction=efconstruction, M=m, random_seed=seed)

        # Add items - position in embeddings is used as the id
        self.backend.add_items(embeddings, np.arange(embeddings.shape[0], dtype=np.int64))

        # Add id offset, delete counter and index build metadata
        self.config["offset"] = embeddings.shape[0]
        self.config["deletes"] = 0
        self.metadata({"efconstruction": efconstruction, "m": m, "seed": seed})

    def append(self, embeddings):
        new = embeddings.shape[0]

        # Resize index
        self.backend.resize_index(self.config["offset"] + new)

        # Append new ids - position in embeddings + existing offset is used as the id
        self.backend.add_items(embeddings, np.arange(self.config["offset"], self.config["offset"] + new, dtype=np.int64))

        # Update id offset and index metadata
        self.config["offset"] += new
        self.metadata()

    def delete(self, ids):
        # Mark elements as deleted to omit from search results
        for uid in ids:
            try:
                self.backend.mark_deleted(uid)
                self.config["deletes"] += 1
            except RuntimeError:
                # Ignore label not found error
                continue

    def search(self, queries, limit):
        # Set ef query param
        ef = self.setting("efsearch")
        if ef:
            self.backend.set_ef(ef)

        # Run the query
        ids, distances = self.backend.knn_query(queries, k=limit)

        # Map results to [(id, score)]
        results = []
        for x, distance in enumerate(distances):
            # Convert distances to similarity scores
            scores = [1 - d for d in distance.tolist()]

            # Build (id, score) tuples, convert np.int64 to python int
            results.append(list(zip(ids[x].tolist(), scores)))

        return results

    def count(self):
        return self.backend.get_current_count() - self.config["deletes"]

    def save(self, path):
        # Write index
        self.backend.save_index(path)



"""
Factory module
"""

from ..util import Resolver

from .hub import HuggingFaceHub
from .storage import ObjectStorage, LIBCLOUD


class CloudFactory:
    """
    Methods to create Cloud instances.
    """

    @staticmethod
    def create(config):
        """
        Creates a Cloud instance.

        Args:
            config: cloud configuration

        Returns:
            Cloud
        """

        # Cloud instance
        cloud = None

        provider = config.get("provider", "")

        # Hugging Face Hub
        if provider.lower() == "huggingface-hub":
            cloud = HuggingFaceHub(config)

        # Cloud object storage
        elif ObjectStorage.isprovider(provider):
            cloud = ObjectStorage(config)

        # External provider
        elif provider:
            cloud = CloudFactory.resolve(provider, config)

        return cloud

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom cloud backend.

        Args:
            backend: backend class
            config: configuration parameters

        Returns:
            Cloud
        """

        try:
            return Resolver()(backend)(config)

        except Exception as e:
            # Failure message
            message = f'Unable to resolve cloud backend: "{backend}".'

            # Append message if LIBCLOUD is not installed
            message += ' Cloud storage is not available - install "cloud" extra to enable' if not LIBCLOUD else ""

            raise ImportError(message) from e



"""
Query module
"""

import logging
import re

try:
    from grandcypher import GrandCypher

    GRANDCYPHER = True
except ImportError:
    GRANDCYPHER = False

# Logging configuration
logger = logging.getLogger(__name__)


class Query:
    """
    Runs openCypher graph queries using the GrandCypher library. This class also supports search functions.
    """

    # Similar token
    SIMILAR = "__SIMILAR__"

    def __init__(self):
        """
        Create a new graph query instance.
        """

        if not GRANDCYPHER:
            raise ImportError('GrandCypher is not available - install "graph" extra to enable')

    def __call__(self, graph, query, limit):
        """
        Runs a graph query.

        Args:
            graph: graph instance
            query: graph query, can be a full query string or a parsed query dictionary
            limit: number of results

        Returns:
            results
        """

        # Results by attribute and ids filter
        attributes, uids = None, None

        # Build the query from a parsed query
        if isinstance(query, dict):
            query, attributes, uids = self.build(query)

        # Filter graph, if applicable
        if uids:
            graph = self.filter(graph, attributes, uids)

        # Debug log graph query
        logger.debug(query)

        # Run openCypher query
        return GrandCypher(graph.backend, limit if limit else 3).run(query)

    def isquery(self, queries):
        """
        Checks a list of queries to see if all queries are openCypher queries.

        Args:
            queries: list of queries to check

        Returns:
            True if all queries are openCypher queries
        """

        # Check for required graph query clauses
        return all(query and query.strip().startswith("MATCH ") and "RETURN " in query for query in queries)

    def parse(self, query):
        """
        Parses a graph query. This method supports parsing search functions and replacing them with placeholders.

        Args:
            query: graph query

        Returns:
            parsed query as a dictionary
        """

        # Parameters
        where, limit, nodes, similar = None, None, [], []

        # Parse where clause
        match = re.search(r"where(.+?)return", query, flags=re.DOTALL | re.IGNORECASE)
        if match:
            where = match.group(1).strip()

        # Parse limit clause
        match = re.search(r"limit\s+(\d+)", query, flags=re.DOTALL | re.IGNORECASE)
        if match:
            limit = match.group(1)

        # Parse similar clauses
        for x, match in enumerate(re.finditer(r"similar\((.+?)\)", query, flags=re.DOTALL | re.IGNORECASE)):
            # Replace similar clause with placeholder
            query = query.replace(match.group(0), f"{Query.SIMILAR}{x}")

            # Parse similar clause parameters
            params = [param.strip().replace("'", "").replace('"', "") for param in match.group(1).split(",")]
            nodes.append(params[0])
            similar.append(params[1:])

        # Return parsed query
        return {
            "query": query,
            "where": where,
            "limit": limit,
            "nodes": nodes,
            "similar": similar,
        }

    def build(self, parse):
        """
        Constructs a full query from a parsed query. This method supports substituting placeholders with search results.

        Args:
            parse: parsed query

        Returns:
            graph query
        """

        # Get query. Initialize attributes and uids.
        query, attributes, uids = parse["query"], {}, {}

        # Replace similar clause with id query
        if "results" in parse:
            for x, result in enumerate(parse["results"]):
                # Get query node
                node = parse["nodes"][x]

                # Add similar match attribute
                attribute = f"match_{x}"
                clause = f"{node}.{attribute} > 0"

                # Replace placeholder with earch results
                query = query.replace(f"{Query.SIMILAR}{x}", f"{clause}")

                # Add uids and scores
                for uid, score in result:
                    if uid not in uids:
                        uids[uid] = score

                # Add results by attribute matched
                attributes[attribute] = result

        # Return query, results by attribute matched and ids filter
        return query, attributes, uids.items()

    def filter(self, graph, attributes, uids):
        """
        Filters the input graph by uids. This method also adds similar match attributes.

        Args:
            graph: graph instance
            attributes: results by attribute matched
            uids: single list with all matching ids

        Returns:
            filtered graph
        """

        # Filter the graph
        graph = graph.filter(uids)

        # Add similar match attributes
        for attribute, result in attributes.items():
            for uid, score in result:
                graph.addattribute(uid, attribute, score)

        return graph



"""
TextToSpeech module
"""

# Conditional import
try:
    import onnxruntime as ort
    import soundfile as sf

    from ttstokenizer import IPATokenizer, TTSTokenizer

    from .signal import Signal, SCIPY

    TTS = SCIPY
except ImportError:
    TTS = False

import json
import logging

from io import BytesIO

import torch
import yaml

import numpy as np

from huggingface_hub.errors import HFValidationError
from transformers import SpeechT5Processor
from transformers.utils import cached_file

from ..base import Pipeline

# Logging configuration
logger = logging.getLogger(__name__)


class TextToSpeech(Pipeline):
    """
    Generates speech from text.
    """

    def __init__(self, path=None, maxtokens=512, rate=22050):
        """
        Creates a new TextToSpeech pipeline.

        Args:
            path: optional model path
            maxtokens: maximum number of tokens model can process, defaults to 512
            rate: target sample rate, defaults to 22050
        """

        if not TTS:
            raise ImportError('TextToSpeech pipeline is not available - install "pipeline" extra to enable')

        # Default path
        path = path if path else "neuml/ljspeech-jets-onnx"

        # Target sample rate
        self.rate = rate

        # Load target tts pipeline
        self.pipeline = None
        if self.hasfile(path, "model.onnx") and self.hasfile(path, "config.yaml"):
            self.pipeline = ESPnet(path, maxtokens, self.providers())
        elif self.hasfile(path, "model.onnx") and self.hasfile(path, "voices.json"):
            self.pipeline = Kokoro(path, maxtokens, self.providers())
        else:
            self.pipeline = SpeechT5(path, maxtokens, self.providers())

    def __call__(self, text, stream=False, speaker=1, encoding=None, **kwargs):
        """
        Generates speech from text. Text longer than maxtokens will be batched and returned
        as a single waveform per text input.

        This method supports text as a string or a list. If the input is a string,
        the return type is audio. If text is a list, the return type is a list.

        Args:
            text: text|list
            stream: stream response if True, defaults to False
            speaker: speaker id, defaults to 1
            encoding: optional audio encoding format
            kwargs: additional keyword args

        Returns:
            list of (audio, sample rate) or list of audio depending on encoding parameter
        """

        # Convert results to a list if necessary
        texts = [text] if isinstance(text, str) else text

        # Streaming response
        if stream:
            return self.stream(texts, speaker, encoding)

        # Transform text to speech
        results = [self.execute(x, speaker, encoding, **kwargs) for x in texts]

        # Return results
        return results[0] if isinstance(text, str) else results

    def providers(self):
        """
        Returns a list of available and usable providers.

        Returns:
            list of available and usable providers
        """

        # Create list of providers, prefer CUDA provider if available
        # CUDA provider only available if GPU is available and onnxruntime-gpu installed
        if torch.cuda.is_available() and "CUDAExecutionProvider" in ort.get_available_providers():
            return [("CUDAExecutionProvider", {"cudnn_conv_algo_search": "DEFAULT"}), "CPUExecutionProvider"]

        # Default when CUDA provider isn't available
        return ["CPUExecutionProvider"]

    def hasfile(self, path, name):
        """
        Tests if a file exists in a local or remote repo.

        Args:
            path: model path
            name: file name

        Returns:
            True if name exists in path, False otherwise
        """

        exists = False
        try:
            # Check if file exists
            exists = cached_file(path_or_repo_id=path, filename=name) is not None
        except (HFValidationError, OSError):
            return False

        return exists

    def stream(self, texts, speaker, encoding):
        """
        Iterates over texts, splits into segments and yields snippets of audio.
        This method is designed to integrate with streaming LLM generation.

        Args:
            texts: list of input texts
            speaker: speaker id
            encoding: audio encoding format

        Returns:
            snippets of audio as NumPy arrays or audio bytes depending on encoding parameter
        """

        buffer = []
        for x in texts:
            buffer.append(x)

            if x == "\n" or (x.strip().endswith(".") and len([y for y in buffer if y]) > 2):
                data, buffer = "".join(buffer), []
                yield self.execute(data, speaker, encoding)

        if buffer:
            data = "".join(buffer)
            yield self.execute(data, speaker, encoding)

    def execute(self, text, speaker, encoding, **kwargs):
        """
        Executes model run for an input array of tokens. This method will build batches
        of tokens when len(tokens) > maxtokens.

        Args:
            text: text to tokenize and pass to model
            speaker: speaker id
            encoding: audio encoding format
            kwargs: additional keyword args

        Returns:
            (audio, sample rate) or audio bytes depending on encoding parameter
        """

        # Run pipeline model
        audio, rate = self.pipeline(text, speaker, **kwargs)

        # Resample, if necessary and return
        audio, rate = (Signal.resample(audio, rate, self.rate), self.rate) if self.rate else (audio, rate)

        # Encoding audio data
        if encoding:
            data = BytesIO()
            sf.write(data, audio, rate, format=encoding)
            return data.getvalue()

        # Default to (audio, rate) tuple
        return (audio, rate)


class SpeechPipeline(Pipeline):
    """
    Base class for speech pipelines
    """

    # pylint: disable=W0221
    def chunk(self, data, size, punctids):
        """
        Batching method that takes punctuation into account. This method splits data up to size
        chunks. But it also searches the batch and splits on the last punctuation token id.

        Args:
            data: data
            size: batch size
            punctids: list of punctuation token ids

        Returns:
            yields batches of data
        """

        # Iterate over each token
        punct, index = 0, 0
        for i, x in enumerate(data):
            # Check if token is a punctuation token
            if x in punctids:
                punct = i

            # Batch size reached, leave a spot for the punctuation token
            if i - index >= (size - 1):
                end = (punct if punct > index else i) + 1
                yield data[index:end]
                index = end

        # Last batch
        if index < len(data):
            yield data[index : len(data)]


class ESPnet(SpeechPipeline):
    """
    Text to Speech pipeline with an ESPnet ONNX model.
    """

    def __init__(self, path, maxtokens, providers):
        """
        Creates a new ESPnet pipeline.

        Args:
            path: model path
            maxtokens: maximum number of tokens model can process
            providers: list of supported ONNX providers
        """

        # Get path to model and config
        config = cached_file(path_or_repo_id=path, filename="config.yaml")
        model = cached_file(path_or_repo_id=path, filename="model.onnx")

        # Read yaml config
        with open(config, "r", encoding="utf-8") as f:
            config = yaml.safe_load(f)

        # Create tokenizer
        tokens = config.get("token", {}).get("list")
        self.tokenizer = TTSTokenizer(tokens)

        # Create ONNX Session
        self.model = ort.InferenceSession(model, ort.SessionOptions(), providers)

        # Max number of input tokens model can handle
        self.maxtokens = maxtokens

        # Get model input name, typically "text"
        self.input = self.model.get_inputs()[0].name

        # Get parameter names
        self.params = set(x.name for x in self.model.get_inputs())

    def __call__(self, text, speaker):
        """
        Executes a model run. This method will build batches of tokens when len(tokens) > maxtokens.

        Args:
            text: text to tokenize and pass to model
            speaker: speaker id

        Returns:
            (audio, sample rate)
        """

        # Debug logging for input text
        logger.debug("%s", text)

        # Sample rate
        rate = 22050

        # Tokenize input
        tokens = self.tokenizer(text)

        # Split into batches and process
        results = []
        for i, x in enumerate(self.chunk(tokens, self.maxtokens, self.tokenizer.punctuation())):
            # Format input parameters
            params = {self.input: x}
            params = {**params, **{"sids": np.array([speaker])}} if "sids" in self.params else params

            # Run text through TTS model and save waveform
            output = self.model.run(None, params)
            results.append(Signal.trim(output[0], rate, trailing=False) if i > 0 else output[0])

        # Concatenate results and return
        return (np.concatenate(results), rate)


class Kokoro(SpeechPipeline):
    """
    Text to Speech pipeline with an Kokoro ONNX model.
    """

    def __init__(self, path, maxtokens, providers):
        """
        Creates a new Kokoro pipeline.

        Args:
            path: model path
            maxtokens: maximum number of tokens model can process
            providers: list of supported ONNX providers
        """

        # Get path to model and config
        voices = cached_file(path_or_repo_id=path, filename="voices.json")
        model = cached_file(path_or_repo_id=path, filename="model.onnx")

        # Read voices config
        with open(voices, "r", encoding="utf-8") as f:
            self.voices = json.load(f)

        # Create tokenizer
        self.tokenizer = IPATokenizer()

        # Create ONNX Session
        self.model = ort.InferenceSession(model, ort.SessionOptions(), providers)

        # Max number of input tokens model can handle
        self.maxtokens = min(maxtokens, 510)

        # Get model input name
        self.input = self.model.get_inputs()[0].name

        # Get parameter names
        self.params = set(x.name for x in self.model.get_inputs())

    def __call__(self, text, speaker=None, speed=1.0, transcribe=True):
        """
        Executes a model run. This method will build batches of tokens when len(tokens) > maxtokens.

        Args:
            text: text to tokenize and pass to model
            speaker: speaker id, defaults to first speaker
            speed: defaults to 1.0
            transcribe: if text should be transcriped to IPA text, defaults to True

        Returns:
            (audio, sample rate)
        """

        # Debug logging for input text
        logger.debug("%s", text)

        # Sample rate
        rate = 24000

        # Looks up speaker, falls back to default
        speaker = speaker if speaker in self.voices else next(iter(self.voices))
        speaker = np.array(self.voices[speaker], dtype=np.float32)

        # Tokenize input
        self.tokenizer.transcribe = transcribe
        tokens = self.tokenizer(text)

        # Split into batches and process
        results = []
        for i, x in enumerate(self.chunk(tokens, self.maxtokens, self.tokenizer.punctuation())):
            # Format input parameters
            params = {self.input: [[0, *x, 0]], "style": speaker[len(x)], "speed": np.ones(1, dtype=np.float32) * speed}

            # Run text through TTS model and save waveform
            output = self.model.run(None, params)
            results.append(Signal.trim(output[0], rate, trailing=False) if i > 0 else output[0])

        # Concatenate results and return
        return (np.concatenate(results), rate)


class SpeechT5(SpeechPipeline):
    """
    Text to Speech pipeline with a SpeechT5 ONNX model.
    """

    def __init__(self, path, maxtokens, providers):
        """
        Creates a new SpeechT5 pipeline.

        Args:
            path: model path
            maxtokens: maximum number of tokens model can process
            providers: list of supported ONNX providers
        """

        self.encoder = ort.InferenceSession(cached_file(path_or_repo_id=path, filename="encoder_model.onnx"), providers=providers)
        self.decoder = ort.InferenceSession(cached_file(path_or_repo_id=path, filename="decoder_model_merged.onnx"), providers=providers)
        self.vocoder = ort.InferenceSession(cached_file(path_or_repo_id=path, filename="decoder_postnet_and_vocoder.onnx"), providers=providers)

        self.processor = SpeechT5Processor.from_pretrained(path)
        self.defaultspeaker = np.load(cached_file(path_or_repo_id=path, filename="speaker.npy"), allow_pickle=False)

        # Max number of input tokens model can handle
        self.maxtokens = maxtokens

        # pylint: disable=E1101
        # Punctuation token ids
        self.punctids = [v for k, v in self.processor.tokenizer.get_vocab().items() if k in ".,!?;"]

    def __call__(self, text, speaker):
        """
        Executes a model run. This method will build batches of tokens when len(tokens) > maxtokens.

        Args:
            text: text to tokenize and pass to model
            speaker: speaker embeddings

        Returns:
            (audio, sample rate)
        """

        # Debug logging for input text
        logger.debug("%s", text)

        # Sample rate
        rate = 16000

        # Tokenize text
        inputs = self.processor(text=text, return_tensors="np", normalize=True)

        # Split into batches and process
        results = []
        for i, x in enumerate(self.chunk(inputs["input_ids"][0], self.maxtokens, self.punctids)):
            # Run text through TTS model and save waveform
            chunk = self.process(np.array([x], dtype=np.int64), speaker)
            results.append(Signal.trim(chunk, rate, trailing=False) if i > 0 else chunk)

        # Concatenate results and return
        return (np.concatenate(results), rate)

    def process(self, inputs, speaker):
        """
        Runs model inference.

        Args:
            inputs: input token ids
            speaker: speaker embeddings

        Returns:
            waveform as NumPy array
        """

        # Run through encoder model
        outputs = self.encoder.run(None, {"input_ids": inputs})
        outputs = {key.name: outputs[x] for x, key in enumerate(self.encoder.get_outputs())}

        # Encoder outputs and parameters
        hiddenstate, attentionmask = outputs["encoder_outputs"], outputs["encoder_attention_mask"]
        minlenratio, maxlenratio = 0.0, 20.0
        reduction, threshold, melbins = 2, 0.5, 80

        maxlen = int(hiddenstate.shape[1] * maxlenratio / reduction)
        minlen = int(hiddenstate.shape[1] * minlenratio / reduction)

        # Main processing loop
        spectrogram, index, crossattention, branch, outputs = [], 0, None, False, {}
        while True:
            index += 1

            inputs = {
                "use_cache_branch": np.array([branch]),
                "encoder_attention_mask": attentionmask,
                "speaker_embeddings": speaker if speaker is not None and isinstance(speaker, np.ndarray) else self.defaultspeaker,
            }

            if index == 1:
                inputs = self.placeholders(inputs)
                inputs["output_sequence"] = np.zeros((1, 1, melbins)).astype(np.float32)
                inputs["encoder_hidden_states"] = hiddenstate
                branch = True
            else:
                inputs = self.inputs(inputs, outputs, crossattention)
                inputs["output_sequence"] = outputs["output_sequence_out"]
                inputs["encoder_hidden_states"] = np.zeros((1, 0, 768)).astype(np.float32)

            # Run inputs through decoder
            outputs = self.decoder.run(None, inputs)
            outputs = {key.name: outputs[x] for x, key in enumerate(self.decoder.get_outputs())}

            # Get cross attention with 1st pass
            if index == 1:
                crossattention = {key: val for key, val in outputs.items() if ("encoder" in key and "present" in key)}

            # Decoder outputs
            prob = outputs["prob"]
            spectrum = outputs["spectrum"]
            spectrogram.append(spectrum)

            # Done when stop token or maximum length is reached.
            if index >= minlen and (int(sum(prob >= threshold)) > 0 or index >= maxlen):
                spectrogram = np.concatenate(spectrogram)
                return self.vocoder.run(None, {"spectrogram": spectrogram})[0]

    def placeholders(self, inputs):
        """
        Creates decoder model inputs for initial inference pass.

        Args:
            inputs: current decoder inputs

        Returns:
            updated decoder inputs
        """

        length = inputs["encoder_attention_mask"].shape[1]

        for x in range(6):
            inputs[f"past_key_values.{x}.encoder.key"] = np.zeros((1, 12, length, 64)).astype(np.float32)
            inputs[f"past_key_values.{x}.encoder.value"] = np.zeros((1, 12, length, 64)).astype(np.float32)
            inputs[f"past_key_values.{x}.decoder.key"] = np.zeros((1, 12, 1, 64)).astype(np.float32)
            inputs[f"past_key_values.{x}.decoder.value"] = np.zeros((1, 12, 1, 64)).astype(np.float32)

        return inputs

    def inputs(self, inputs, previous, crossattention):
        """
        Creates decoder model inputs for follow-on inference passes.

        Args:
            inputs: current decoder inputs
            previous: previous decoder outputs
            crossattention: crossattention parameters

        Returns:
            updated decoder inputs
        """

        for x in range(6):
            inputs[f"past_key_values.{x}.encoder.key"] = crossattention[f"present.{x}.encoder.key"]
            inputs[f"past_key_values.{x}.encoder.value"] = crossattention[f"present.{x}.encoder.value"]
            inputs[f"past_key_values.{x}.decoder.key"] = previous[f"present.{x}.decoder.key"]
            inputs[f"past_key_values.{x}.decoder.value"] = previous[f"present.{x}.decoder.value"]

        return inputs



"""
Function imports
"""

from smolagents import Tool


class FunctionTool(Tool):
    """
    Creates a FunctionTool. A FunctionTool takes descriptive configuration and injects it along with a target function
    into an LLM prompt.
    """

    # pylint: disable=W0231
    def __init__(self, config):
        """
        Creates a FunctionTool.

        Args:
            config: `name`, `description`, `inputs`, `output` and `target` configuration
        """

        # Tool parameters
        self.name = config["name"]
        self.description = config["description"]
        self.inputs = config["inputs"]
        self.output_type = config.get("output", config.get("output_type", "any"))
        self.target = config["target"]

        # pylint: disable=C0103
        # Skip forward signature validation
        self.skip_forward_signature_validation = True

        # Validate parameters and initialize tool
        super().__init__()

    def forward(self, *args, **kwargs):
        """
        Runs target function.

        Args:
            args: positional args
            kwargs: keyword args

        Returns:
            result
        """

        return self.target(*args, **kwargs)



"""
Scan module
"""


class Scan:
    """
    Scans indexes for query matches.
    """

    def __init__(self, search, limit, weights, index):
        """
        Creates a new scan instance.

        Args:
            search: index search function
            limit: maximum results
            weights: default hybrid score weights
            index: default index name
        """

        # Index search function
        self.search = search

        # Default query limit
        self.limit = limit

        # Default number of candidates
        self.candidates = None

        # Default query weights
        self.weights = weights

        # Default index
        self.index = index

    def __call__(self, queries, parameters):
        """
        Executes a scan for a list of queries.

        Args:
            queries: list of queries to run
            parameters: list of dicts of named parameters to bind to placeholders

        Returns:
            list of (id, score) per query
        """

        # Query results group by unique query clause id
        results = {}

        # Default number of candidates
        default = None

        # Group by index and run
        for index, iqueries in self.parse(queries, parameters).items():
            # Query limit to pass to batch search
            candidates = [query.candidates for query in iqueries if query.candidates]
            if not candidates and not default:
                default = self.default(queries)

            candidates = max(candidates) if candidates else default

            # Query weights to pass to batch search
            weights = [query.weights for query in iqueries if query.weights is not None]
            weights = max(weights) if weights else self.weights

            # Index to run query against
            index = index if index else self.index

            # Run index searches
            for x, result in enumerate(self.search([query.text for query in iqueries], candidates, weights, index)):
                # Save query id and results to later join to original query
                results[iqueries[x].uid] = (iqueries[x].qid, result)

        # Sort by query uid and return results
        return [result for _, result in sorted(results.items())]

    def parse(self, queries, parameters):
        """
        Parse index query clauses from a list of parsed queries.

        Args:
            queries: list of parsed queries
            parameters: list of dicts of named parameters to bind to placeholders

        Returns:
            index query clauses grouped by index
        """

        results, uid = {}, 0
        for x, query in enumerate(queries):
            if "similar" in query:
                # Extract similar query clauses
                for params in query["similar"]:
                    # Resolve bind parameters
                    if parameters and parameters[x]:
                        params = self.bind(params, parameters[x])

                    # Parse query clause
                    clause = Clause(uid, x, params)

                    # Create clause list for index
                    if clause.index not in results:
                        results[clause.index] = []

                    # Add query to index list, increment uid
                    results[clause.index].append(clause)
                    uid += 1

        return results

    def bind(self, similar, parameters):
        """
        Resolves bind parameters for a similar function call.

        Args:
            similar: similar function call arguments
            parameters: bind parameters

        Returns:
            similar function call arguments with resolved bind parameters
        """

        resolved = []
        for p in similar:
            # Resolve bind parameters
            if isinstance(p, str) and p.startswith(":") and p[1:] in parameters:
                resolved.append(parameters[p[1:]])
            else:
                resolved.append(p)

        return resolved

    def default(self, queries):
        """
        Derives the default number of candidates. The number of candidates are the number of results to bring back
        from index queries. This is an optional argument to similar() clauses.

        For a single query filter clause, the default is the query limit. With multiple filtering clauses, the default is
        10x the query limit. This ensures that limit results are still returned with additional filtering after an index query.

        Args:
            queries: list of queries

        Returns:
            default candidate list size
        """

        multitoken = any(query.get("where") and len(query["where"].split()) > 1 for query in queries)
        return self.limit * 10 if multitoken else self.limit


class Clause:
    """
    Parses and stores query clause parameters.
    """

    def __init__(self, uid, qid, params):
        """
        Creates a new query clause.

        Args:
            uid: query clause id
            qid: query id clause is a part of
            params: query parameters to parse
        """

        self.uid, self.qid = uid, qid
        self.text, self.index = params[0], None
        self.candidates, self.weights = None, None

        # Parse additional similar clause parameters
        if len(params) > 1:
            self.parse(params[1:])

    def parse(self, params):
        """
        Parses clause parameters into this instance.

        Args:
            params: query clause parameters
        """

        for param in params:
            if (isinstance(param, str) and param.isdigit()) or isinstance(param, int):
                # Number of query candidates
                self.candidates = int(param)

            elif (isinstance(param, str) and param.replace(".", "").isdigit()) or isinstance(param, float):
                # Hybrid score weights
                self.weights = float(param)

            else:
                # Target index
                self.index = param



"""
Resolver module
"""


class Resolver:
    """
    Resolves a Python class path
    """

    def __call__(self, path):
        """
        Class instance to resolve.

        Args:
            path: path to class

        Returns:
            class instance
        """

        # Split into path components
        parts = path.split(".")

        # Resolve each path component
        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)

        # Return class instance
        return m



"""
Reducer module
"""

from zipfile import BadZipFile

# Conditionally import dimensionality reduction libraries as they aren't installed by default
try:
    import skops.io as sio

    from sklearn.decomposition import TruncatedSVD

    REDUCER = True
except ImportError:
    REDUCER = False

from ...serialize import SerializeFactory


class Reducer:
    """
    LSA dimensionality reduction model
    """

    def __init__(self, embeddings=None, components=None):
        """
        Creates a dimensionality reduction model.

        Args:
            embeddings: input embeddings matrix
            components: number of model components
        """

        if not REDUCER:
            raise ImportError('Dimensionality reduction is not available - install "vectors" extra to enable')

        self.model = self.build(embeddings, components) if embeddings is not None and components else None

    def __call__(self, embeddings):
        """
        Applies a dimensionality reduction model to embeddings, removed the top n principal components. Operation applied
        directly on array.

        Args:
            embeddings: input embeddings matrix
        """

        pc = self.model.components_
        factor = embeddings.dot(pc.transpose())

        # Apply LSA model
        # Calculation is different if n_components = 1
        if pc.shape[0] == 1:
            embeddings -= factor * pc
        elif len(embeddings.shape) > 1:
            # Apply model on a row-wise basis to limit memory usage
            for x in range(embeddings.shape[0]):
                embeddings[x] -= factor[x].dot(pc)
        else:
            # Single embedding
            embeddings -= factor.dot(pc)

    def build(self, embeddings, components):
        """
        Builds a LSA model. This model is used to remove the principal component within embeddings. This helps to
        smooth out noisy embeddings (common words with less value).

        Args:
            embeddings: input embeddings matrix
            components: number of model components

        Returns:
            LSA model
        """

        model = TruncatedSVD(n_components=components, random_state=0)
        model.fit(embeddings)

        return model

    def load(self, path):
        """
        Loads a Reducer object from path.

        Args:
            path: directory path to load model
        """

        # Dimensionality reduction
        try:
            self.model = sio.load(path)
        except (BadZipFile, KeyError):
            # Backwards compatible support for pickled models
            self.model = SerializeFactory.create("pickle").load(path)

    def save(self, path):
        """
        Saves a Reducer object to path.

        Args:
            path: directory path to save model
        """

        sio.dump(self.model, path)



"""
Pipeline imports
"""

from .audio import *
from .base import Pipeline
from .data import *
from .factory import PipelineFactory
from .hfmodel import HFModel
from .hfpipeline import HFPipeline
from .image import *
from .llm import *
from .llm import RAG as Extractor
from .nop import Nop
from .text import *
from .tensors import Tensors
from .train import *



"""
Sparse module
"""

from queue import Queue
from threading import Thread

from ..ann import SparseANNFactory
from ..vectors import SparseVectorsFactory

from .base import Scoring


class Sparse(Scoring):
    """
    Sparse vector scoring.
    """

    # End of stream message
    COMPLETE = 1

    def __init__(self, config=None, models=None):
        super().__init__(config)

        # Vector configuration
        mapping = {"vectormethod": "method", "vectornormalize": "normalize"}
        config = {k: v for k, v in config.items() if k not in mapping.values()}
        for k, v in mapping.items():
            if k in config:
                config[v] = config[k]

        # Load the SparseVectors model
        self.model = SparseVectorsFactory.create(config, models)

        # Normalize search outputs if vectors are not normalized already
        # A float can also be provided to set the normalization factor (defaults to 30.0)
        self.isnormalize = self.config.get("normalize", True)

        # Sparse ANN
        self.ann = None

        # Encoding processing parameters
        self.batch = self.config.get("batch", 1024)
        self.thread, self.queue, self.data = None, None, None

    def insert(self, documents, index=None, checkpoint=None):
        # Start processing thread, if necessary
        self.start(checkpoint)

        data = []
        for uid, document, tags in documents:
            # Extract text, if necessary
            if isinstance(document, dict):
                document = document.get(self.text, document.get(self.object))

            if document is not None:
                # Add data
                data.append((uid, " ".join(document) if isinstance(document, list) else document, tags))

        # Add batch of data
        self.queue.put(data)

    def delete(self, ids):
        self.ann.delete(ids)

    def index(self, documents=None):
        # Insert documents, if provided
        if documents:
            self.insert(documents)

        # Create ANN, if there is pending data
        embeddings = self.stop()
        if embeddings is not None:
            self.ann = SparseANNFactory.create(self.config)
            self.ann.index(embeddings)

    def upsert(self, documents=None):
        # Insert documents, if provided
        if documents:
            self.insert(documents)

        # Check for existing index and pending data
        if self.ann:
            embeddings = self.stop()
            if embeddings is not None:
                self.ann.append(embeddings)
        else:
            self.index()

    def weights(self, tokens):
        # Not supported
        return None

    def search(self, query, limit=3):
        return self.batchsearch([query], limit)[0]

    def batchsearch(self, queries, limit=3, threads=True):
        # Convert queries to embedding vectors
        embeddings = self.model.batchtransform((None, query, None) for query in queries)

        # Run ANN search
        scores = self.ann.search(embeddings, limit)

        # Normalize scores if normalization IS enabled AND vector normalization IS NOT enabled
        return self.normalize(embeddings, scores) if self.isnormalize and not self.model.isnormalize else scores

    def count(self):
        return self.ann.count()

    def load(self, path):
        self.ann = SparseANNFactory.create(self.config)
        self.ann.load(path)

    def save(self, path):
        # Save Sparse ANN
        if self.ann:
            self.ann.save(path)

    def close(self):
        # Close Sparse ANN
        if self.ann:
            self.ann.close()

        # Clear parameters
        self.model, self.ann, self.thread, self.queue = None, None, None, None

    def issparse(self):
        return True

    def isnormalized(self):
        return self.isnormalize or self.model.isnormalize

    def start(self, checkpoint):
        """
        Starts an encoding processing thread.

        Args:
            checkpoint: checkpoint directory
        """

        if not self.thread:
            self.queue = Queue(5)
            self.thread = Thread(target=self.encode, args=(checkpoint,))
            self.thread.start()

    def stop(self):
        """
        Stops an encoding processing thread. Return processed results.

        Returns:
            results
        """

        results = None
        if self.thread:
            # Send EOS message
            self.queue.put(Sparse.COMPLETE)

            self.thread.join()
            self.thread, self.queue = None, None

            # Get return value
            results = self.data
            self.data = None

        return results

    def encode(self, checkpoint):
        """
        Encodes streaming data.

        Args:
            checkpoint: checkpoint directory
        """

        # Streaming encoding of data
        _, dimensions, self.data = self.model.vectors(self.stream(), self.batch, checkpoint)

        # Save number of dimensions
        self.config["dimensions"] = dimensions

    def stream(self):
        """
        Streams data from an input queue until end of stream message received.
        """

        batch = self.queue.get()
        while batch != Sparse.COMPLETE:
            yield from batch
            batch = self.queue.get()

    def normalize(self, queries, scores):
        """
        Normalize query result using the max query score.

        Args:
            queries: query vectors
            scores: query results

        Returns:
            normalized query results
        """

        # Get normalize scale factor
        scale = 30.0 if isinstance(self.isnormalize, bool) else self.isnormalize

        # Normalize scores using max scores
        maxscores = self.model.dot(queries, queries)

        # Normalize results and return
        results = []
        for x, result in enumerate(scores):
            maxscore = max(maxscores[x][x] / scale, scale)
            maxscore = max(maxscore, result[0][1]) if result else maxscore

            results.append([(uid, score / maxscore) for uid, score in result])

        return results



"""
LLM imports
"""

from .factory import GenerationFactory
from .generation import Generation
from .huggingface import *
from .litellm import LiteLLM
from .llama import LlamaCpp
from .llm import LLM
from .rag import RAG



"""
FileTask module
"""

import os
import re

from .base import Task


class FileTask(Task):
    """
    Task that processes file paths
    """

    # File prefix
    FILE = r"file:\/\/"

    def accept(self, element):
        # Replace file prefixes
        element = re.sub(FileTask.FILE, "", element)

        # Only accept file paths that exist
        return super().accept(element) and isinstance(element, str) and os.path.exists(element)

    def prepare(self, element):
        # Replace file prefixes
        return re.sub(FileTask.FILE, "", element)



"""
Defines API paths for TTS endpoints
"""

from typing import Optional

from fastapi import APIRouter, Response

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/texttospeech")
def texttospeech(text: str, speaker: Optional[str] = None, encoding: Optional[str] = "mp3"):
    """
    Generates speech from text.

    Args:
        text: text
        speaker: speaker id, defaults to 1
        encoding: optional audio encoding format

    Returns:
        Audio data
    """

    # Convert to audio
    audio = application.get().pipeline("texttospeech", text, speaker=speaker, encoding=encoding)

    # Write audio
    return Response(audio, headers={"Content-Disposition": f"attachment;filename=speech.{encoding.lower()}"})



"""
Embeddings module
"""

from smolagents import Tool

from ...embeddings import Embeddings


class EmbeddingsTool(Tool):
    """
    Tool to execute an Embeddings search.
    """

    def __init__(self, config):
        """
        Creates a new EmbeddingsTool.

        Args:
            config: embeddings tool configuration
        """

        # Tool parameters
        self.name = config["name"]
        self.description = f"""{config['description']}. Results are returned as a list of dict elements.
Each result has keys 'id', 'text', 'score'."""

        # Input and output descriptions
        self.inputs = {"query": {"type": "string", "description": "The search query to perform."}}
        self.output_type = "any"

        # Load embeddings instance
        self.embeddings = self.load(config)

        # Validate parameters and initialize tool
        super().__init__()

    # pylint: disable=W0221
    def forward(self, query):
        """
        Runs a search.

        Args:
            query: input query

        Returns:
            search results
        """

        return self.embeddings.search(query, 5)

    def load(self, config):
        """
        Loads an embeddings instance from config.

        Args:
            config: embeddings tool configuration

        Returns:
            Embeddings
        """

        if "target" in config:
            return config["target"]

        embeddings = Embeddings()
        embeddings.load(**config)

        return embeddings



"""
Factory module
"""

from ...util import Resolver

from .external import External
from .huggingface import HFVectors
from .litellm import LiteLLM
from .llama import LlamaCpp
from .m2v import Model2Vec
from .sbert import STVectors
from .words import WordVectors


class VectorsFactory:
    """
    Methods to create dense vector models.
    """

    @staticmethod
    def create(config, scoring=None, models=None):
        """
        Create a Vectors model instance.

        Args:
            config: vector configuration
            scoring: scoring instance
            models: models cache

        Returns:
            Vectors
        """

        # Determine vector method
        method = VectorsFactory.method(config)

        # External vectors
        if method == "external":
            return External(config, scoring, models)

        # LiteLLM vectors
        if method == "litellm":
            return LiteLLM(config, scoring, models)

        # llama.cpp vectors
        if method == "llama.cpp":
            return LlamaCpp(config, scoring, models)

        # Model2vec vectors
        if method == "model2vec":
            return Model2Vec(config, scoring, models)

        # Sentence Transformers vectors
        if method == "sentence-transformers":
            return STVectors(config, scoring, models) if config and config.get("path") else None

        # Word vectors
        if method == "words":
            return WordVectors(config, scoring, models)

        # Transformers vectors
        if HFVectors.ismethod(method):
            return HFVectors(config, scoring, models) if config and config.get("path") else None

        # Resolve custom method
        return VectorsFactory.resolve(method, config, scoring, models) if method else None

    @staticmethod
    def method(config):
        """
        Get or derive the vector method.

        Args:
            config: vector configuration

        Returns:
            vector method
        """

        # Determine vector method
        method = config.get("method")
        path = config.get("path")

        # Infer method from path, if blank
        if not method:
            if path:
                if LiteLLM.ismodel(path):
                    method = "litellm"
                elif LlamaCpp.ismodel(path):
                    method = "llama.cpp"
                elif Model2Vec.ismodel(path):
                    method = "model2vec"
                elif WordVectors.ismodel(path):
                    method = "words"
                else:
                    method = "transformers"
            elif config.get("transform"):
                method = "external"

        return method

    @staticmethod
    def resolve(backend, config, scoring, models):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: vector configuration
            scoring: scoring instance
            models: models cache

        Returns:
            Vectors
        """

        try:
            return Resolver()(backend)(config, scoring, models)
        except Exception as e:
            raise ImportError(f"Unable to resolve vectors backend: '{backend}'") from e



"""
Defines API paths for textractor endpoints.
"""

from typing import List

from fastapi import APIRouter, Body

from .. import application
from ..route import EncodingAPIRoute

router = APIRouter(route_class=EncodingAPIRoute)


@router.get("/textract")
def textract(file: str):
    """
    Extracts text from a file at path.

    Args:
        file: file to extract text

    Returns:
        extracted text
    """

    return application.get().pipeline("textractor", (file,))


@router.post("/batchtextract")
def batchtextract(files: List[str] = Body(...)):
    """
    Extracts text from a file at path.

    Args:
        files: list of files to extract text

    Returns:
        list of extracted text
    """

    return application.get().pipeline("textractor", (files,))



"""
Factory module
"""

from .base import Archive


class ArchiveFactory:
    """
    Methods to create Archive instances.
    """

    @staticmethod
    def create(directory=None):
        """
        Create a new Archive instance.

        Args:
            directory: optional default working directory, otherwise uses a temporary directory

        Returns:
            Archive
        """

        return Archive(directory)
